# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from gi.repository import Gst

from xl.providers import ProviderHandler

import logging

logger = logging.getLogger(__name__)


def element_link_many(*elems):
    for i in range(len(elems) - 1):
        e = elems[i]
        n = elems[i + 1]
        e.link(n)


def disable_video_text(bin):
    # From quod libet: disable all video/text decoding in playbin
    GST_PLAY_FLAG_VIDEO = 1 << 0
    GST_PLAY_FLAG_TEXT = 1 << 2
    flags = bin.get_property("flags")
    flags &= ~(GST_PLAY_FLAG_VIDEO | GST_PLAY_FLAG_TEXT)
    bin.set_property("flags", flags)


class ElementBin(Gst.Bin):
    """
    A bin for easily containing elements

    elements are added to the elements dictionary in the form of
        elements[position] = element
    where position is a value from 0-100 indicating its position
    in the resulting bin, and element is the Gst.Element itself.

    changes made to elements do not apply until setup_elements()
    is called
    """

    def __init__(self, name=None):
        if name:
            Gst.Bin.__init__(self, name=name)
        else:
            Gst.Bin.__init__(self)

        self.elements = {}
        self.added_elems = []
        self.srcpad = None
        self.sinkpad = None
        self.src = None
        self.sink = None

    def setup_elements(self):
        """
        This function should only be called when state is NULL

        :returns: True if there are elements, False otherwise
        """
        name = self.get_name()
        logger.debug("%s: Removing all elements", name)

        if len(self.added_elems) > 0:
            for elem in self.added_elems:
                elem.set_state(Gst.State.NULL)
                self.remove(elem)

        all_elems = sorted(self.elements.items())

        elems = []
        for _unused, e in all_elems:
            # Don't add empty elements!
            if hasattr(e, 'setup_elements'):
                if not e.setup_elements():
                    continue

            elems.append(e)

        self.added_elems = elems

        if len(elems) == 0:
            return False

        for e in elems:
            self.add(e)
            e.sync_state_with_parent()
            logger.debug("%s: Adding %s", name, e.get_name())

        element_link_many(*elems)

        self.srcpad = elems[-1].get_static_pad("src")
        if self.src is not None:
            self.src.set_target(self.srcpad)
        else:
            self.src = Gst.GhostPad.new('src', self.srcpad)
            self.add_pad(self.src)

        self.sinkpad = elems[0].get_static_pad("sink")
        if self.sink is not None:
            self.sink.set_target(self.sinkpad)
        else:
            self.sink = Gst.GhostPad.new('sink', self.sinkpad)
            self.add_pad(self.sink)

        return True


class ProviderBin(ElementBin, ProviderHandler):
    """
    A ProviderBin is a Gst.Bin that adds and removes elements from itself
    using the providers system. Providers should be a subclass of
    Gst.Element and provide the following attributes:
        name  - name to use for this element
        index - priority within the pipeline. range [0-100] integer.
                lower numbers are higher priority. elements must
                choose a unique number.
    """

    def __init__(self, servicename, name=None):
        """
        :param servicename: the Provider name to listen for
        """
        if name is None:
            name = servicename
        ElementBin.__init__(self, name=name)
        ProviderHandler.__init__(self, servicename)

        self.reset_providers()

    def reset_providers(self):
        self.elements = {}
        dups = {}
        for provider in self.get_providers():
            idx = provider.index
            if idx in self.elements:
                dup = dups.setdefault(idx, [self.elements[idx].name])
                dup.append(provider.name)
                while idx in self.elements:
                    idx += 1
            try:
                self.elements[idx] = provider()
            except Exception:
                logger.exception(
                    "Could not create %s element for %s.", provider, self.get_name()
                )

        for k, v in dups.items():
            logger.warning(
                "Audio plugins %s are sharing index %s (may have unpredictable output!)",
                v,
                k,
            )

    def on_provider_added(self, provider):
        self.reset_providers()

    def on_provider_removed(self, provider):
        self.reset_providers()


def parse_stream_tags(track, tag_list):
    """
    Called when a tag is found in a stream.

    :type track:    xl.trax.Track
    :type tag_list: Gst.TagList

    Gst.TagList guarantees that all strings are either ASCII or UTF8
    """

    newsong = False

    keep = [
        'bitrate',
        'duration',
        'track-number',
        'track-count',
        'album-disc-number',
        'album-disc-count',
        'album',
        'artist',
        'genre',
        'comment',
        'title',
        'datetime',
    ]

    # Build a dictionary first
    tags = {}
    for i in range(tag_list.n_tags()):
        k = tag_list.nth_tag_name(i)
        if k not in keep:
            continue

        values = [
            tag_list.get_value_index(k, vi) for vi in range(tag_list.get_tag_size(k))
        ]

        tags[k] = values

    etags = {}

    v = tags.get('bitrate')
    if v:
        etags['__bitrate'] = int(v[0])

    v = tags.get('duration')
    if v:
        etags['__length'] = float(v[0]) / Gst.SECOND

    v = tags.get('track-number')
    if v:
        c = tags.get('track-count')
        if c:
            etags['tracknumber'] = '%d/%d' % (v[0], c[0])
        else:
            etags['tracknumber'] = '%d' % (v[0])

    v = tags.get('album-disc-number')
    if v:
        c = tags.get('album-disc-count')
        if c:
            etags['discnumber'] = '%d/%d' % (v[0], c[0])
        else:
            etags['discnumber'] = '%d' % (v[0])

    v = tags.get('album')
    if v:
        etags['album'] = v

    v = tags.get('artist')
    if v:
        etags['artist'] = v

    v = tags.get('genre')
    if v:
        etags['genre'] = v

    v = tags.get('datetime')
    if v:
        # v[0] is a Gst.DateTime object
        etags['date'] = v[0].to_iso8601_string()

    # if there's a comment, but no album, set album to the comment
    v = tags.get('comment')
    if v and not track.get_tag_raw('album'):
        etags['album'] = v

    v = tags.get('title')
    if v:
        try:
            if track.get_tag_raw('__rawtitle') != v:
                etags['__rawtitle'] = v
                newsong = True
        except AttributeError:
            etags['__rawtitle'] = v
            newsong = True

        if not etags.get('artist'):
            title_array = v[0].split(' - ', 1)
            if len(title_array) == 1 or track.get_loc_for_io().lower().endswith(".mp3"):
                etags['title'] = v
            else:
                etags['artist'] = [title_array[0]]
                etags['title'] = [title_array[1]]
        else:
            etags['title'] = v

    track.set_tags(**etags)
    return newsong


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import os.path
import sys

try:
    uid = os.geteuid()
except Exception:
    pass
else:
    if uid == 0:
        sys.stderr.write('Error: running as root is not supported!\n')

if sys.platform == 'linux':
    # Set process name
    try:
        import ctypes

        libc = ctypes.CDLL('libc.so.6')
        libc.prctl(15, b'exaile', 0, 0, 0)  # 15 = PR_SET_NAME
    except Exception:
        pass

# Set visible process name. Requires module "setproctitle"
try:
    from setproctitle import setproctitle

    setproctitle('exaile')
except ImportError:
    pass

# Find the location of exaile's working directory and insert it to sys.path
basedir = os.path.dirname(os.path.realpath(__file__))
if not os.path.exists(os.path.join(basedir, "exaile.py")):
    cwd = os.getcwd()
    if os.path.exists(os.path.join(cwd, "exaile.py")):
        basedir = cwd
sys.path.insert(0, basedir)
os.environ['EXAILE_DIR'] = basedir


def main():
    from xl import main

    global exaile
    exaile = main.Exaile()


if __name__ == "__main__":
    main()

# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


from xl.metadata._base import CaseInsensitiveBaseFormat
from mutagen import apev2


class ApeFormat(CaseInsensitiveBaseFormat):
    MutagenType = apev2.APEv2
    # TODO: this mapping is incomplete
    tag_mapping = {
        "title": "Title",
        "album": "Album",
        "artist": "Artist",
        "albumartist": "Album Artist",
        "genre": "Genre",
        "tracknumber": "Track",
        "date": "Year",
        "composer": "Composer",
        "discnumber": "Disc",
        "comment": "Comment",
        "isrc": "ISRC",
        "lyrics": "Lyrics",
        "albumsort": "ALBUMSORT",
        "albumartistsort": "ALBUMARTISTSORT",
        "artistsort": "ARTISTSORT",
        "titlesort": "TITLESORT",
        "grouping": "Grouping",
        "language": "Language",
    }
    others = True
    writable = True


# vim: et sts=4 sw=4



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging
import os
import os.path
import tempfile
import threading

import cairo
from gi.repository import Gio
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from xl import common, event, providers, settings, xdg
from xl.covers import MANAGER as COVER_MANAGER
from xl.nls import gettext as _
from xlgui.widgets import dialogs, menu
from xlgui import guiutil
from xlgui.guiutil import pixbuf_from_data

logger = logging.getLogger(__name__)


def save_pixbuf(pixbuf, path, type_):
    """Save a pixbuf to a local file.

    :param pixbuf: Pixbuf to save
    :type pixbuf: GdkPixbuf.Pixbuf
    :param path: Path of file to save to
    :type path: str
    :param type_: Type of image file. See GdkPixbuf.savev for valid values.
    :type type_: str
    :return: None
    """
    # This wraps the horrible GdkPixbuf.savev API. Can be removed if one day
    # PyGObject provides an override.
    pixbuf.savev(path, type_, [None], [])


class CoverManager(GObject.Object):
    """
    Cover manager window
    """

    __gsignals__ = {
        'prefetch-started': (GObject.SignalFlags.RUN_LAST, None, ()),
        'prefetch-progress': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,)),
        'prefetch-completed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,)),
        'fetch-started': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,)),
        'fetch-completed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,)),
        'fetch-progress': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_INT,)),
        'cover-fetched': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT, GdkPixbuf.Pixbuf),
        ),
    }

    def __init__(self, parent, collection):
        """
        Initializes the window
        """
        GObject.Object.__init__(self)

        # List of identifiers of albums without covers
        self.outstanding = []
        # Map of album identifiers and their tracks
        self.album_tracks = {}

        self.outstanding_text = _('{outstanding} covers left to fetch')
        self.completed_text = _('All covers fetched')
        self.cover_size = (90, 90)
        self.default_cover_pixbuf = pixbuf_from_data(
            COVER_MANAGER.get_default_cover(), self.cover_size
        )

        builder = guiutil.get_builder(xdg.get_data_path('ui', 'covermanager.ui'))

        builder.connect_signals(self)

        self.window = builder.get_object('window')
        self.window.set_transient_for(parent)

        self.message = dialogs.MessageBar(
            parent=builder.get_object('content_area'), buttons=Gtk.ButtonsType.CLOSE
        )

        self.previews_box = builder.get_object('previews_box')
        self.model = builder.get_object('covers_model')
        # Map of album identifiers and model paths
        self.model_path_cache = {}
        self.menu = CoverMenu(self)
        self.menu.attach_to_widget(self.previews_box, lambda menu, widget: True)

        self.progress_bar = builder.get_object('progressbar')
        self.progress_bar.set_text(_('Collecting albums and covers...'))
        self.progress_bar.pulse_timeout = GLib.timeout_add(
            100, self.on_progress_pulse_timeout
        )
        self.close_button = builder.get_object('close_button')
        self.stop_button = builder.get_object('stop_button')
        self.stop_button.set_sensitive(False)
        self.fetch_button = builder.get_object('fetch_button')

        self.window.show_all()

        self.stopper = threading.Event()
        thread = threading.Thread(
            target=self.prefetch, name='CoverPrefetch', args=(collection,)
        )
        thread.daemon = True
        thread.start()

    def prefetch(self, collection):
        """
        Collects all albums and sets the list of outstanding items
        """
        albums = set()

        for track in collection:
            if self.stopper.is_set():
                return

            try:
                artist = track.get_tag_raw('artist')[0]
                album = track.get_tag_raw('album')[0]
            except TypeError:
                continue

            if not album or not artist:
                continue

            album = (artist, album)

            try:
                self.album_tracks[album].append(track)
            except KeyError:
                self.album_tracks[album] = [track]

            albums.add(album)

        albums = sorted(albums)

        outstanding = []
        # Speed up the following loop
        get_cover = COVER_MANAGER.get_cover
        default_cover_pixbuf = self.default_cover_pixbuf
        cover_size = self.cover_size

        self.emit('prefetch-started')

        for i, album in enumerate(albums):
            if self.stopper.is_set():
                return

            cover_data = get_cover(self.album_tracks[album][0], set_only=True)
            cover_pixbuf = pixbuf_from_data(cover_data) if cover_data else None

            try:
                thumbnail_pixbuf = cover_pixbuf.scale_simple(
                    *cover_size, interp_type=GdkPixbuf.InterpType.BILINEAR
                )
            except AttributeError:  # cover_pixbuf is None
                thumbnail_pixbuf = default_cover_pixbuf
                outstanding.append(album)

            label = '{0} - {1}'.format(*album)
            iter = self.model.append((album, thumbnail_pixbuf, label))
            self.model_path_cache[album] = self.model.get_path(iter)

            self.emit('prefetch-progress', i + 1)

        self.outstanding = outstanding
        self.emit('prefetch-completed', len(self.outstanding))

    def fetch(self):
        """
        Collects covers for all outstanding items
        """
        self.emit('fetch-started', len(self.outstanding))

        # Speed up the following loop
        get_cover = COVER_MANAGER.get_cover
        save = COVER_MANAGER.save

        for i, album in enumerate(self.outstanding[:]):
            if self.stopper.is_set():
                # Allow for "fetch-completed" signal to be emitted
                break

            cover_data = get_cover(self.album_tracks[album][0], save_cover=True)
            cover_pixbuf = pixbuf_from_data(cover_data) if cover_data else None

            self.emit('fetch-progress', i + 1)

            if not cover_pixbuf:
                continue

            self.outstanding.remove(album)
            self.emit('cover-fetched', album, cover_pixbuf)

            if i % 50 == 0:
                logger.debug('Saving cover database')
                save()

        logger.debug('Saving cover database')
        save()

        self.emit('fetch-completed', len(self.outstanding))

    def show_cover(self):
        """
        Shows the currently selected cover
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            path = paths[0]
            album = self.model[path][0]
            track = self.album_tracks[album][0]  # Arbitrary track in album
            cover_data = COVER_MANAGER.get_cover(track, set_only=True)
            cover_pixbuf = pixbuf_from_data(cover_data) if cover_data else None

            # Do not bother showing the dialog if there is no cover
            if cover_pixbuf:
                savedir = Gio.File.new_for_uri(track.get_loc_for_io()).get_parent()
                if savedir:
                    savedir = savedir.get_path()
                cover_window = CoverWindow(self.window, cover_pixbuf, album[1], savedir)
                cover_window.show_all()

    def fetch_cover(self):
        """
        Shows the cover chooser for the currently selected album
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            path = paths[0]
            album = self.model[path][0]
            track = self.album_tracks[album][0]
            cover_chooser = CoverChooser(self.window, track)
            # Make sure we're updating the correct album after selection
            cover_chooser.path = path
            cover_chooser.connect('cover-chosen', self.on_cover_chosen)

    def remove_cover(self):
        """
        Removes the cover of the currently selected album
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            path = paths[0]
            album = self.model[path][0]
            track = self.album_tracks[album][0]
            COVER_MANAGER.remove_cover(track)
            self.model[path][1] = self.default_cover_pixbuf

    @common.idle_add()
    def do_prefetch_started(self):
        """
        Sets the widget states to prefetching
        """
        self.previews_box.set_model(None)
        self.model.clear()
        self.previews_box.set_sensitive(False)
        self.fetch_button.set_sensitive(False)
        self.progress_bar.set_fraction(0)
        GLib.source_remove(self.progress_bar.pulse_timeout)

    @common.idle_add()
    def do_prefetch_completed(self, outstanding):
        """
        Sets the widget states to ready for fetching
        """
        self.previews_box.set_sensitive(True)
        self.previews_box.set_model(self.model)
        self.fetch_button.set_sensitive(True)
        self.progress_bar.set_fraction(0)
        self.progress_bar.set_text(
            self.outstanding_text.format(outstanding=outstanding)
        )

    @common.idle_add()
    def do_prefetch_progress(self, progress):
        """
        Updates the wiedgets to reflect the processed album
        """
        fraction = progress / float(len(self.album_tracks))
        self.progress_bar.set_fraction(fraction)

    @common.idle_add()
    def do_fetch_started(self, outstanding):
        """
        Sets the widget states to fetching
        """
        self.previews_box.set_sensitive(False)
        self.stop_button.set_sensitive(True)
        self.fetch_button.set_sensitive(False)
        self.progress_bar.set_fraction(0)
        # We need float for the fraction during progress
        self.progress_bar.outstanding_total = float(outstanding)

    @common.idle_add()
    def do_fetch_completed(self, outstanding):
        """
        Sets the widget states to ready for fetching
        """
        self.previews_box.set_sensitive(True)
        self.stop_button.set_sensitive(False)

        if outstanding > 0:
            # If there are covers left for some reason, allow re-fetch
            self.fetch_button.set_sensitive(True)

        self.progress_bar.set_fraction(0)

    @common.idle_add()
    def do_fetch_progress(self, progress):
        """
        Updates the widgets to reflect the processed album
        """
        outstanding = len(self.outstanding)

        if outstanding > 0:
            progress_text = self.outstanding_text.format(outstanding=outstanding)
        else:
            progress_text = self.completed_text

        self.progress_bar.set_text(progress_text)

        fraction = progress / self.progress_bar.outstanding_total
        self.progress_bar.set_fraction(fraction)

    @common.idle_add()
    def do_cover_fetched(self, album, pixbuf):
        """
        Updates the widgets to reflect the newly fetched cover
        """
        path = self.model_path_cache[album]
        self.model[path][1] = pixbuf.scale_simple(
            *self.cover_size, interp_type=GdkPixbuf.InterpType.BILINEAR
        )

    def on_cover_chosen(self, cover_chooser, track, cover_data):
        """
        Updates the cover of the current album after user selection
        """
        path = cover_chooser.path

        if path:
            album = self.model[path][0]
            pixbuf = pixbuf_from_data(cover_data)

            self.emit('cover-fetched', album, pixbuf)

            try:
                self.outstanding.remove(album)
            except ValueError:
                pass
            else:
                outstanding = len(self.outstanding)

                if outstanding > 0:
                    progress_text = self.outstanding_text.format(
                        outstanding=outstanding
                    )
                else:
                    progress_text = self.completed_text

                self.progress_bar.set_text(progress_text)

    def on_previews_box_item_activated(self, iconview, path):
        """
        Shows the currently selected cover
        """
        self.show_cover()

    def on_previews_box_button_press_event(self, widget, e):
        """
        Shows the cover menu upon click
        """
        path = self.previews_box.get_path_at_pos(int(e.x), int(e.y))

        if path:
            self.previews_box.select_path(path)

            if e.triggers_context_menu():
                self.menu.popup(None, None, None, None, 3, e.time)

    def on_previews_box_popup_menu(self, menu):
        """
        Shows the cover menu upon keyboard interaction
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            self.menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def on_previews_box_query_tooltip(self, widget, x, y, keyboard_mode, tooltip):
        """
        Custom tooltip display to prevent markup errors
        (e.g. due to album names containing "<")
        """
        x, y = self.previews_box.convert_widget_to_bin_window_coords(x, y)
        path = self.previews_box.get_path_at_pos(x, y)

        if path:
            tooltip.set_text(self.model[path][2])
            self.previews_box.set_tooltip_item(tooltip, path)

            return True

        return False

    def on_progress_pulse_timeout(self):
        """
        Updates the progress during prefetching
        """
        self.progress_bar.pulse()

        return True

    def on_close_button_clicked(self, button):
        """
        Stops the current fetching process and closes the dialog
        """
        self.stopper.set()
        self.window.destroy()

        # Free some memory
        self.model.clear()
        del self.outstanding
        del self.album_tracks
        del self.model_path_cache

    def on_stop_button_clicked(self, button):
        """
        Stops the current fetching process
        """
        self.stopper.set()

    def on_fetch_button_clicked(self, button):
        """
        Starts the cover fetching process
        """
        self.stopper.clear()
        thread = threading.Thread(target=self.fetch, name='CoverFetch')
        thread.daemon = True
        thread.start()

    def on_window_delete_event(self, window, e):
        """
        Stops the current fetching process and closes the dialog
        """
        self.close_button.clicked()

        return True


class CoverMenu(menu.Menu):
    """
    Cover menu
    """

    def __init__(self, widget):
        """
        Initializes the menu
        """
        menu.Menu.__init__(self, widget)
        self.w = widget

        self.add_simple(_('Show Cover'), self.on_show_clicked)
        self.add_simple(_('Fetch Cover'), self.on_fetch_clicked)
        self.add_simple(_('Remove Cover'), self.on_remove_clicked)

    def on_show_clicked(self, *e):
        """
        Shows the current cover
        """
        self.w.show_cover()

    def on_fetch_clicked(self, *e):
        self.w.fetch_cover()

    def on_remove_clicked(self, *e):
        self.w.remove_cover()


class CoverWidget(Gtk.EventBox):
    """
    Represents the cover widget displayed by the track information
    """

    __gsignals__ = {'cover-found': (GObject.SignalFlags.RUN_LAST, None, (object,))}

    def __init__(self, image):
        """
        Initializes the widget

        :param image: the image to wrap
        :type image: :class:`Gtk.Image`
        """
        GObject.Object.__init__(self)

        self.image = image
        self.cover_data = None
        self.menu = CoverMenu(self)
        self.menu.attach_to_widget(self)

        self.filename = None

        guiutil.gtk_widget_replace(image, self)
        self.add(self.image)
        self.set_track(None)
        self.image.show()

        event.add_callback(self.on_quit_application, 'quit_application')

        if settings.get_option('gui/use_alpha', False):
            self.set_app_paintable(True)

    def destroy(self):
        """
        Cleanups
        """
        if self.filename is not None and os.path.exists(self.filename):
            os.remove(self.filename)
            self.filename = None

        event.remove_callback(self.on_quit_application, 'quit-application')

    def set_track(self, track):
        """
        Fetches album covers, and displays them
        """

        self.__track = track

        self.set_blank()
        self.drag_dest_set(
            Gtk.DestDefaults.ALL,
            [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
            Gdk.DragAction.COPY | Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE,
        )

        @common.threaded
        def __get_cover():
            fetch = not settings.get_option('covers/automatic_fetching', True)
            cover_data = COVER_MANAGER.get_cover(track, set_only=fetch)

            if not cover_data:
                return

            GLib.idle_add(self.on_cover_chosen, None, track, cover_data)

        if track is not None:
            __get_cover()

    def show_cover(self):
        """
        Shows the current cover
        """
        if not self.cover_data:
            return

        pixbuf = pixbuf_from_data(self.cover_data)

        if pixbuf:
            savedir = Gio.File.new_for_uri(self.__track.get_loc_for_io()).get_parent()
            if savedir:
                savedir = savedir.get_path()
            window = CoverWindow(
                self.get_toplevel(),
                pixbuf,
                self.__track.get_tag_display('album'),
                savedir,
            )
            window.show_all()

    def fetch_cover(self):
        """
        Fetches a cover for the current track
        """
        if not self.__track:
            return

        window = CoverChooser(self.get_toplevel(), self.__track)
        window.connect('cover-chosen', self.on_cover_chosen)

    def remove_cover(self):
        """
        Removes the cover for the current track from the database
        """
        COVER_MANAGER.remove_cover(self.__track)
        self.set_blank()

    def set_blank(self):
        """
        Sets the default cover to display
        """

        self.drag_dest_unset()

        pixbuf = pixbuf_from_data(COVER_MANAGER.get_default_cover())
        self.image.set_from_pixbuf(pixbuf)
        self.set_drag_source_enabled(False)
        self.cover_data = None

        self.emit('cover-found', None)

    def set_drag_source_enabled(self, enabled):
        """
        Changes the behavior for drag and drop

        :param drag_enabled: Whether to  allow
            drag to other applications
        :type enabled: bool
        """
        if enabled == getattr(self, '__drag_source_enabled', None):
            return

        if enabled:
            self.drag_source_set(
                Gdk.ModifierType.BUTTON1_MASK,
                [Gtk.TargetEntry.new('text/uri-list', 0, 0)],
                Gdk.DragAction.DEFAULT | Gdk.DragAction.MOVE,
            )
        else:
            self.drag_source_unset()

        self.__drag_source_enabled = enabled

    def do_button_press_event(self, event):
        """
        Called when someone clicks on the cover widget
        """
        if self.__track is None or self.get_toplevel() is None:
            return

        if event.type == Gdk.EventType._2BUTTON_PRESS:
            self.show_cover()
        elif event.triggers_context_menu():
            self.menu.popup(event)

    def do_expose_event(self, event):
        """
        Paints alpha transparency
        """
        opacity = 1 - settings.get_option('gui/transparency', 0.3)
        context = self.props.window.cairo_create()
        background = self.style.bg[Gtk.StateType.NORMAL]
        context.set_source_rgba(
            float(background.red) / 256**2,
            float(background.green) / 256**2,
            float(background.blue) / 256**2,
            opacity,
        )
        context.set_operator(cairo.OPERATOR_SOURCE)
        context.paint()

        Gtk.EventBox.do_expose_event(self, event)

    def do_drag_begin(self, context):
        """
        Sets the cover as drag icon
        """
        self.drag_source_set_icon_pixbuf(self.image.get_pixbuf())

    def do_drag_data_get(self, context, selection, info, time):
        """
        Fills the selection with the current cover
        """
        if self.filename is None:
            self.filename = tempfile.mkstemp(prefix='exaile_cover_')[1]

        pixbuf = pixbuf_from_data(self.cover_data)
        save_pixbuf(pixbuf, self.filename, 'png')
        selection.set_uris([Gio.File.new_for_path(self.filename).get_uri()])

    def do_drag_data_delete(self, context):
        """
        Cleans up after drag from cover widget
        """
        if self.filename is not None and os.path.exists(self.filename):
            os.remove(self.filename)
            self.filename = None

    def do_drag_data_received(self, context, x, y, selection, info, time):
        """
        Sets the cover based on the dragged data
        """
        if self.__track is not None:
            uri = selection.get_uris()[0]
            db_string = 'localfile:%s' % uri

            try:
                stream = Gio.File.new_for_uri(uri).read()
            except GLib.Error:
                return

            self.cover_data = stream.read()
            width = settings.get_option('gui/cover_width', 100)
            pixbuf = pixbuf_from_data(self.cover_data, (width, width))

            if pixbuf is not None:
                self.image.set_from_pixbuf(pixbuf)
                COVER_MANAGER.set_cover(self.__track, db_string, self.cover_data)

    def on_cover_chosen(self, object, track, cover_data):
        """
        Called when a cover is selected
        from the coverchooser
        """

        if self.__track != track:
            return

        width = settings.get_option('gui/cover_width', 100)
        pixbuf = pixbuf_from_data(cover_data, (width, width))
        self.image.set_from_pixbuf(pixbuf)
        self.set_drag_source_enabled(True)
        self.cover_data = cover_data

        self.emit('cover-found', pixbuf)

    def on_track_tags_changed(self, e, track, tags):
        """
        Updates the displayed cover upon tag changes
        """
        if self.__track == track:
            cover_data = COVER_MANAGER.get_cover(track)

            if not cover_data:
                return

            GLib.idle_add(self.on_cover_chosen, None, cover_data)

    def on_quit_application(self, type, exaile, nothing):
        """
        Cleans up temporary files
        """
        if self.filename is not None and os.path.exists(self.filename):
            os.remove(self.filename)
            self.filename = None


class CoverWindow:
    """Shows the cover in a simple image viewer"""

    def __init__(self, parent, pixbuf, album=None, savedir=None):
        """Initializes and shows the cover

        :param parent: Parent window to attach to
        :type parent: Gtk.Window
        :param pixbuf: Pixbuf of the cover image
        :type pixbuf: GdkPixbuf.Pixbuf
        :param album: Album title
        :type album: basestring
        :param savedir: Initial directory for the Save As functionality
        :type savedir: basestring
        """
        self.builder = guiutil.get_builder(xdg.get_data_path('ui', 'coverwindow.ui'))
        self.builder.connect_signals(self)

        self.cover_window = self.builder.get_object('CoverWindow')
        self.layout = self.builder.get_object('layout')
        self.toolbar = self.builder.get_object('toolbar')
        self.save_as_button = self.builder.get_object('save_as_button')
        self.zoom_in_button = self.builder.get_object('zoom_in_button')
        self.zoom_out_button = self.builder.get_object('zoom_out_button')
        self.zoom_100_button = self.builder.get_object('zoom_100_button')
        self.zoom_fit_button = self.builder.get_object('zoom_fit_button')
        self.close_button = self.builder.get_object('close_button')
        self.image = self.builder.get_object('image')
        self.statusbar = self.builder.get_object('statusbar')
        self.scrolledwindow = self.builder.get_object('scrolledwindow')
        self.scrolledwindow.set_hadjustment(self.layout.get_hadjustment())
        self.scrolledwindow.set_vadjustment(self.layout.get_vadjustment())

        if album:
            title = _('Cover for %s') % album
        else:
            title = _('Cover')
        self.savedir = savedir

        self.cover_window.set_title(title)
        self.cover_window.set_transient_for(parent)
        self.cover_window_width = 500
        tb_min_height, tb_natural_height = self.toolbar.get_preferred_height()
        sb_min_height, sb_natural_height = self.statusbar.get_preferred_height()
        self.cover_window_height = 500 + tb_natural_height + sb_natural_height
        self.cover_window.set_default_size(
            self.cover_window_width, self.cover_window_height
        )

        self.image_original_pixbuf = pixbuf
        self.image_pixbuf = self.image_original_pixbuf
        self.min_percent = 1
        self.max_percent = 500
        self.ratio = 1.5
        self.image_interp = GdkPixbuf.InterpType.BILINEAR
        self.image_fitted = True
        self.set_ratio_to_fit()
        self.update_widgets()

    def show_all(self):
        self.cover_window.show_all()

    def available_image_width(self):
        """Returns the available horizontal space for the image"""
        return self.cover_window.get_size()[0]

    def available_image_height(self):
        """Returns the available vertical space for the image"""
        tb_min_height, tb_natural_height = self.toolbar.get_preferred_height()
        sb_min_height, sb_natural_height = self.statusbar.get_preferred_height()

        return self.cover_window.get_size()[1] - tb_natural_height - sb_natural_height

    def center_image(self):
        """Centers the image in the layout"""
        new_x = max(
            0, (self.available_image_width() - self.image_pixbuf.get_width()) // 2
        )
        new_y = max(
            0, (self.available_image_height() - self.image_pixbuf.get_height()) // 2
        )
        self.layout.move(self.image, new_x, new_y)

    def update_widgets(self):
        """Updates image, layout, scrolled window, tool bar and status bar"""
        window = self.cover_window.get_window()
        if window:
            window.freeze_updates()
        self.apply_zoom()
        self.layout.set_size(
            self.image_pixbuf.get_width(), self.image_pixbuf.get_height()
        )
        if self.image_fitted or (
            self.image_pixbuf.get_width() == self.available_image_width()
            and self.image_pixbuf.get_height() == self.available_image_height()
        ):
            self.scrolledwindow.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)
        else:
            self.scrolledwindow.set_policy(
                Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
            )
        percent = int(100 * self.image_ratio)
        message = _("{width}x{height} pixels ({zoom}%)").format(
            width=self.image_original_pixbuf.get_width(),
            height=self.image_original_pixbuf.get_height(),
            zoom=percent,
        )
        self.zoom_in_button.set_sensitive(percent < self.max_percent)
        self.zoom_out_button.set_sensitive(percent > self.min_percent)
        self.statusbar.pop(self.statusbar.get_context_id(''))
        self.statusbar.push(self.statusbar.get_context_id(''), message)
        self.image.set_from_pixbuf(self.image_pixbuf)
        self.center_image()
        if window:
            window.thaw_updates()

    def apply_zoom(self):
        """Scales the image if needed"""
        new_width = int(self.image_original_pixbuf.get_width() * self.image_ratio)
        new_height = int(self.image_original_pixbuf.get_height() * self.image_ratio)
        if (
            new_width != self.image_pixbuf.get_width()
            or new_height != self.image_pixbuf.get_height()
        ):
            self.image_pixbuf = self.image_original_pixbuf.scale_simple(
                new_width, new_height, self.image_interp
            )

    def set_ratio_to_fit(self):
        """Calculates and sets the needed ratio to show the full image"""
        width_ratio = (
            float(self.image_original_pixbuf.get_width()) / self.available_image_width()
        )
        height_ratio = (
            float(self.image_original_pixbuf.get_height())
            / self.available_image_height()
        )
        self.image_ratio = 1 / max(1, width_ratio, height_ratio)

    def on_key_press(self, widget, event, data=None):
        """
        Closes the cover window when Escape or Ctrl+W is pressed
        """
        if event.keyval == Gdk.KEY_Escape or (
            event.state & Gdk.ModifierType.CONTROL_MASK and event.keyval == Gdk.KEY_w
        ):
            widget.destroy()

    def on_save_as_button_clicked(self, widget):
        """
        Saves image to user-specified location
        """
        dialog = Gtk.FileChooserDialog(
            _("Save File"),
            self.cover_window,
            Gtk.FileChooserAction.SAVE,
            (
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.CANCEL,
                Gtk.STOCK_SAVE,
                Gtk.ResponseType.ACCEPT,
            ),
        )
        names = settings.get_option('covers/localfile/preferred_names')
        filename = (names[0] if names else 'cover') + '.png'
        dialog.set_current_name(filename)
        if self.savedir:
            dialog.set_current_folder(self.savedir)
        if dialog.run() == Gtk.ResponseType.ACCEPT:
            filename = dialog.get_filename()
            lowfilename = filename.lower()
            if lowfilename.endswith('.jpg') or lowfilename.endswith('.jpeg'):
                type_ = 'jpeg'
            else:
                type_ = 'png'
            save_pixbuf(self.image_pixbuf, filename, type_)
        dialog.destroy()

    def on_zoom_in_button_clicked(self, widget):
        """
        Zooms into the image
        """
        self.image_fitted = False
        self.image_ratio *= self.ratio
        self.update_widgets()

    def on_zoom_out_button_clicked(self, widget):
        """
        Zooms out of the image
        """
        self.image_fitted = False
        self.image_ratio *= 1 / self.ratio
        self.update_widgets()

    def on_zoom_100_button_clicked(self, widget):
        """
        Restores the original image zoom
        """
        self.image_fitted = False
        self.image_ratio = 1
        self.update_widgets()

    def on_zoom_fit_button_clicked(self, widget):
        """
        Zooms the image to fit the window width
        """
        self.image_fitted = True
        self.set_ratio_to_fit()
        self.update_widgets()

    def on_close_button_clicked(self, widget):
        """
        Hides the window
        """
        self.cover_window.hide()

    def cover_window_size_allocate(self, widget, allocation):
        if (
            self.cover_window_width != allocation.width
            or self.cover_window_height != allocation.height
        ):
            if self.image_fitted:
                self.set_ratio_to_fit()
            self.update_widgets()
            self.cover_window_width = allocation.width
            self.cover_window_height = allocation.height


class CoverChooser(GObject.Object):
    """
    Fetches all album covers for a string, and allows the user to choose
    one out of the list
    """

    __gsignals__ = {
        'covers-fetched': (GObject.SignalFlags.RUN_LAST, None, (object,)),
        'cover-chosen': (GObject.SignalFlags.RUN_LAST, None, (object, object)),
    }

    def __init__(self, parent, track, search=None):
        """
        Expects the parent control, a track, an an optional search string
        """
        GObject.Object.__init__(self)
        self.parent = parent

        self.builder = guiutil.get_builder(xdg.get_data_path('ui', 'coverchooser.ui'))
        self.builder.connect_signals(self)
        self.window = self.builder.get_object('CoverChooser')

        self.window.set_title(
            _("Cover options for %(artist)s - %(album)s")
            % {
                'artist': track.get_tag_display('artist'),
                'album': track.get_tag_display('album'),
            }
        )
        self.window.set_transient_for(parent)

        self.message = dialogs.MessageBar(
            parent=self.builder.get_object('main_container'),
            buttons=Gtk.ButtonsType.CLOSE,
        )
        self.message.connect('response', self.on_message_response)

        self.track = track
        self.covers = []
        self.current = 0

        self.cover = guiutil.ScalableImageWidget()
        self.cover.set_image_size(350, 350)

        self.cover_image_box = self.builder.get_object('cover_image_box')

        self.stack = self.builder.get_object('stack')
        self.stack_ready = self.builder.get_object('stack_ready')

        self.size_label = self.builder.get_object('size_label')
        self.source_label = self.builder.get_object('source_label')

        self.covers_model = self.builder.get_object('covers_model')
        self.previews_box = self.builder.get_object('previews_box')
        self.previews_box.set_no_show_all(True)
        self.previews_box.hide()
        self.previews_box.set_model(None)

        self.set_button = self.builder.get_object('set_button')
        self.set_button.set_sensitive(False)

        self.window.show_all()

        self.stopper = threading.Event()
        self.fetcher_thread = threading.Thread(
            target=self.fetch_cover, name='Coverfetcher'
        )
        self.fetcher_thread.start()

    def fetch_cover(self):
        """
        Searches for covers for the current track
        """
        db_strings = COVER_MANAGER.find_covers(self.track)

        if db_strings:
            for db_string in db_strings:
                if self.stopper.is_set():
                    return

                coverdata = COVER_MANAGER.get_cover_data(db_string)
                # Pre-render everything for faster display later
                pixbuf = pixbuf_from_data(coverdata)

                if pixbuf:
                    self.covers_model.append(
                        [
                            (db_string, coverdata),
                            pixbuf,
                            pixbuf.scale_simple(50, 50, GdkPixbuf.InterpType.BILINEAR),
                        ]
                    )

        self.emit('covers-fetched', db_strings)

    def do_covers_fetched(self, db_strings):
        """
        Finishes the dialog setup after all covers have been fetched
        """
        if self.stopper.is_set():
            return

        self.stack.set_visible_child(self.stack_ready)
        self.previews_box.set_model(self.covers_model)

        if db_strings:
            self.cover_image_box.pack_start(self.cover, True, True, 0)
            self.cover.show()
            self.set_button.set_sensitive(True)

            # Show thumbnail bar if more than one cover was found
            if len(db_strings) > 1:
                self.previews_box.set_no_show_all(False)
                self.previews_box.show_all()

            # Try to select the current cover of the track, fallback to first
            track_db_string = COVER_MANAGER.get_db_string(self.track)
            position = (
                db_strings.index(track_db_string)
                if track_db_string in db_strings
                else 0
            )
            self.previews_box.select_path(Gtk.TreePath(position))
        else:
            self.builder.get_object('stack').hide()
            self.builder.get_object('actions_box').hide()
            self.message.show_warning(
                _('No covers found.'),
                _(
                    'None of the enabled sources has a cover for this track, try enabling more sources.'
                ),
            )

    def on_cancel_button_clicked(self, button):
        """
        Closes the cover chooser
        """
        # Notify the fetcher thread to stop
        self.stopper.set()

        self.window.destroy()

    def on_set_button_clicked(self, button):
        """
        Chooses the current cover and saves it to the database
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            path = paths[0]
            coverdata = self.covers_model[path][0]

            COVER_MANAGER.set_cover(self.track, coverdata[0], coverdata[1])

            self.emit('cover-chosen', self.track, coverdata[1])
            self.window.destroy()

    def on_previews_box_selection_changed(self, iconview):
        """
        Switches the currently displayed cover
        """
        paths = self.previews_box.get_selected_items()

        if paths:
            path = paths[0]
            db_string = self.covers_model[path][0]
            source = db_string[0].split(':', 1)[0]
            provider = providers.get_provider('covers', source)
            pixbuf = self.covers_model[path][1]

            self.cover.set_image_pixbuf(pixbuf)
            self.size_label.set_text(
                _('{width}x{height} pixels').format(
                    width=pixbuf.get_width(), height=pixbuf.get_height()
                )
            )
            # Display readable title of the provider, fallback to its name
            self.source_label.set_text(getattr(provider, 'title', source))

            self.set_button.set_sensitive(True)
        else:
            self.set_button.set_sensitive(False)

    def on_previews_box_item_activated(self, iconview, path):
        """
        Triggers selecting the current cover
        """
        self.set_button.clicked()

    def on_message_response(self, widget, response):
        """
        Handles the response for closing
        """
        if response == Gtk.ResponseType.CLOSE:
            self.window.destroy()



# Copyright (C) 2008-2010 Adam Olsen
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import logging

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from xl.nls import gettext as _
from xl import common, event, formatter, player, providers, settings, trax
from xlgui.accelerators import AcceleratorManager
from xlgui.accelerators import Accelerator
from xlgui.playlist_container import PlaylistContainer
from xlgui.widgets import dialogs, info, menu, playback
from xlgui.widgets.playlist import PlaylistPage, PlaylistView
from xlgui import guiutil, tray, menu as mainmenu

logger = logging.getLogger(__name__)

# Length of playback step when user presses seek key (sec)
SEEK_STEP_DEFAULT = 10

# Length of volume steps when user presses up/down key
VOLUME_STEP_DEFAULT = 0.1


class MainWindow(GObject.Object):
    """
    Main Exaile Window
    """

    __gproperties__ = {
        'is-fullscreen': (
            bool,
            'Fullscreen',
            'Whether the window is fullscreen.',
            False,  # Default
            GObject.ParamFlags.READWRITE,
        )
    }

    __gsignals__ = {'main-visible-toggle': (GObject.SignalFlags.RUN_LAST, bool, ())}

    _mainwindow = None

    def __init__(self, controller, builder, collection):
        """
        Initializes the main window

        @param controller: the main gui controller
        """
        GObject.Object.__init__(self)

        self.controller = controller
        self.collection = collection
        self.playlist_manager = controller.exaile.playlists
        self.current_page = -1
        self._fullscreen = False

        self.window_state = 0
        self.minimized = False

        self.builder = builder

        self.window = self.builder.get_object('ExaileWindow')
        self.window.set_title('Exaile')
        self.title_formatter = formatter.TrackFormatter(
            settings.get_option(
                'gui/main_window_title_format', _('$title (by $artist)') + ' - Exaile'
            )
        )

        self.accel_group = Gtk.AccelGroup()
        self.window.add_accel_group(self.accel_group)
        self.accel_manager = AcceleratorManager(
            'mainwindow-accelerators', self.accel_group
        )
        self.menubar = self.builder.get_object("mainmenu")

        fileitem = self.builder.get_object("file_menu_item")
        filemenu = menu.ProviderMenu('menubar-file-menu', self)
        fileitem.set_submenu(filemenu)

        edititem = self.builder.get_object("edit_menu_item")
        editmenu = menu.ProviderMenu('menubar-edit-menu', self)
        edititem.set_submenu(editmenu)

        viewitem = self.builder.get_object("view_menu_item")
        viewmenu = menu.ProviderMenu('menubar-view-menu', self)
        viewitem.set_submenu(viewmenu)

        toolsitem = self.builder.get_object("tools_menu_item")
        toolsmenu = menu.ProviderMenu('menubar-tools-menu', self)
        toolsitem.set_submenu(toolsmenu)

        helpitem = self.builder.get_object("help_menu_item")
        helpmenu = menu.ProviderMenu('menubar-help-menu', self)
        helpitem.set_submenu(helpmenu)

        self._setup_widgets()
        self._setup_position()
        self._setup_hotkeys()
        logger.info("Connecting main window events...")
        self._connect_events()
        MainWindow._mainwindow = self

        mainmenu._create_menus()

    def _setup_hotkeys(self):
        """
        Sets up accelerators that haven't been set up in UI designer
        """

        def factory(integer, description):
            """Generate key bindings for Alt keys"""
            keybinding = '<Alt>%s' % str(integer)
            callback = lambda *_e: self._on_focus_playlist_tab(integer - 1)
            return (keybinding, description, callback)

        hotkeys = (
            (
                '<Primary>S',
                _('Save currently selected playlist'),
                lambda *_e: self.on_save_playlist(),
            ),
            (
                '<Shift><Primary>S',
                _('Save currently selected playlist under a custom name'),
                lambda *_e: self.on_save_playlist_as(),
            ),
            (
                '<Primary>F',
                _('Focus filter in currently focused panel'),
                lambda *_e: self.on_panel_filter_focus(),
            ),
            (
                '<Primary>G',
                _('Focus playlist search'),
                lambda *_e: self.on_search_playlist_focus(),
            ),  # FIXME
            (
                '<Primary><Alt>l',
                _('Clear queue'),
                lambda *_e: player.QUEUE.clear(),
            ),  # FIXME
            (
                '<Primary>P',
                _('Start, pause or resume the playback'),
                self._on_playpause_button,
            ),
            (
                '<Primary>Right',
                _('Seek to the right'),
                lambda *_e: self._on_seek_key(True),
            ),
            (
                '<Primary>Left',
                _('Seek to the left'),
                lambda *_e: self._on_seek_key(False),
            ),
            (
                '<Primary>plus',
                _('Increase the volume'),
                lambda *_e: self._on_volume_key(True),
            ),
            (
                '<Primary>equal',
                _('Increase the volume'),
                lambda *_e: self._on_volume_key(True),
            ),
            (
                '<Primary>minus',
                _('Decrease the volume'),
                lambda *_e: self._on_volume_key(False),
            ),
            ('<Primary>Page_Up', _('Switch to previous tab'), self._on_prev_tab_key),
            ('<Primary>Page_Down', _('Switch to next tab'), self._on_next_tab_key),
            (
                '<Alt>N',
                _('Focus the playlist container'),
                self._on_focus_playlist_container,
            ),
            # These 4 are subject to change.. probably should do this
            # via a different mechanism too...
            (
                '<Alt>I',
                _('Focus the files panel'),
                lambda *_e: self.controller.focus_panel('files'),
            ),
            # ('<Alt>C', _('Focus the collection panel'),  # TODO: Does not work, why?
            # lambda *_e: self.controller.focus_panel('collection')),
            (
                '<Alt>R',
                _('Focus the radio panel'),
                lambda *_e: self.controller.focus_panel('radio'),
            ),
            (
                '<Alt>L',
                _('Focus the playlists panel'),
                lambda *_e: self.controller.focus_panel('playlists'),
            ),
            factory(1, _('Focus the first tab')),
            factory(2, _('Focus the second tab')),
            factory(3, _('Focus the third tab')),
            factory(4, _('Focus the fourth tab')),
            factory(5, _('Focus the fifth tab')),
            factory(6, _('Focus the sixth tab')),
            factory(7, _('Focus the seventh tab')),
            factory(8, _('Focus the eighth tab')),
            factory(9, _('Focus the ninth tab')),
            factory(0, _('Focus the tenth tab')),
        )

        for keys, helptext, function in hotkeys:
            accelerator = Accelerator(keys, helptext, function)
            providers.register('mainwindow-accelerators', accelerator)

    def _setup_widgets(self):
        """
        Sets up the various widgets
        """
        # TODO: Maybe make this stackable
        self.message = dialogs.MessageBar(
            parent=self.builder.get_object('player_box'), buttons=Gtk.ButtonsType.CLOSE
        )

        self.info_area = MainWindowTrackInfoPane(player.PLAYER)
        self.info_area.set_auto_update(True)
        self.info_area.set_border_width(3)
        self.info_area.hide()
        self.info_area.set_no_show_all(True)
        guiutil.gtk_widget_replace(self.builder.get_object('info_area'), self.info_area)

        self.statusbar = MainWindowStatusBarPane(player.PLAYER)
        guiutil.gtk_widget_replace(
            self.builder.get_object('status_bar'), self.statusbar
        )

        self.volume_control = playback.VolumeControl(player.PLAYER)
        self.info_area.get_action_area().pack_end(self.volume_control, False, False, 0)

        if settings.get_option('gui/use_alpha', False):
            screen = self.window.get_screen()
            visual = screen.get_rgba_visual()
            self.window.set_visual(visual)
            self.window.connect('screen-changed', self.on_screen_changed)
            self._update_alpha()

        self._update_dark_hint()

        playlist_area = self.builder.get_object('playlist_area')
        self.playlist_container = PlaylistContainer('saved_tabs', player.PLAYER)
        for notebook in self.playlist_container.notebooks:
            notebook.connect_after(
                'switch-page', self.on_playlist_container_switch_page
            )
            page = notebook.get_current_tab()
            if page is not None:
                selection = page.view.get_selection()
                selection.connect('changed', self.on_playlist_view_selection_changed)

        playlist_area.pack_start(self.playlist_container, True, True, 3)

        self.splitter = self.builder.get_object('splitter')

        # In most (all?) RTL locales, the playback controls should still be LTR.
        # Just in case that's not always the case, we provide a hidden option to
        # force RTL layout instead. This can be removed once we're more certain
        # that the default behavior (always LTR) is correct.
        controls_direction = (
            Gtk.TextDirection.RTL
            if settings.get_option('gui/rtl_playback_controls')
            else Gtk.TextDirection.LTR
        )

        self.play_image = Gtk.Image.new_from_icon_name(
            'media-playback-start', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.play_image.set_direction(controls_direction)
        self.pause_image = Gtk.Image.new_from_icon_name(
            'media-playback-pause', Gtk.IconSize.SMALL_TOOLBAR
        )
        self.pause_image.set_direction(controls_direction)

        self.stop_image = Gtk.Image.new_from_icon_name(
            'media-playback-stop', Gtk.IconSize.BUTTON
        )
        self.stop_image.set_direction(controls_direction)
        self.spat_image = Gtk.Image.new_from_icon_name(
            'process-stop', Gtk.IconSize.BUTTON
        )
        self.spat_image.set_direction(controls_direction)

        play_toolbar = self.builder.get_object('play_toolbar')
        play_toolbar.set_direction(controls_direction)
        for button in ('playpause', 'next', 'prev', 'stop'):
            widget = self.builder.get_object('%s_button' % button)
            setattr(self, '%s_button' % button, widget)
            widget.get_child().set_direction(controls_direction)

        """Set the button icon size here to prevent resizing later"""
        self.playpause_button.set_image(self.play_image)
        self.stop_button.set_image(self.stop_image)

        self.progress_bar = playback.SeekProgressBar(player.PLAYER)
        self.progress_bar.get_child().set_direction(controls_direction)
        # Don't expand vertically; looks awful on Adwaita.
        self.progress_bar.set_valign(Gtk.Align.CENTER)
        guiutil.gtk_widget_replace(
            self.builder.get_object('playback_progressbar_dummy'), self.progress_bar
        )

        self.stop_button.toggle_spat = False
        self.stop_button.add_events(Gdk.EventMask.POINTER_MOTION_MASK)
        self.stop_button.connect(
            'motion-notify-event', self.on_stop_button_motion_notify_event
        )
        self.stop_button.connect(
            'leave-notify-event', self.on_stop_button_leave_notify_event
        )
        self.stop_button.connect('key-press-event', self.on_stop_button_key_press_event)
        self.stop_button.connect(
            'key-release-event', self.on_stop_button_key_release_event
        )
        self.stop_button.connect('focus-out-event', self.on_stop_button_focus_out_event)
        self.stop_button.connect('button-press-event', self.on_stop_button_press_event)
        self.stop_button.connect(
            'button-release-event', self.on_stop_button_release_event
        )
        self.stop_button.drag_dest_set(
            Gtk.DestDefaults.ALL,
            [Gtk.TargetEntry.new("exaile-index-list", Gtk.TargetFlags.SAME_APP, 0)],
            Gdk.DragAction.COPY,
        )
        self.stop_button.connect('drag-motion', self.on_stop_button_drag_motion)
        self.stop_button.connect('drag-leave', self.on_stop_button_drag_leave)
        self.stop_button.connect(
            'drag-data-received', self.on_stop_button_drag_data_received
        )

        event.add_ui_callback(self.on_exaile_loaded, 'exaile_loaded')

    def _connect_events(self):
        """
        Connects the various events to their handlers
        """
        self.builder.connect_signals(
            {
                'on_configure_event': self.configure_event,
                'on_window_state_event': self.window_state_change_event,
                'on_delete_event': self.on_delete_event,
                'on_playpause_button_clicked': self._on_playpause_button,
                'on_next_button_clicked': lambda *e: player.QUEUE.next(),
                'on_prev_button_clicked': lambda *e: player.QUEUE.prev(),
                'on_about_item_activate': self.on_about_item_activate,
                # Controller
                #            'on_scan_collection_item_activate': self.controller.on_rescan_collection,
                #            'on_device_manager_item_activate': lambda *e: self.controller.show_devices(),
                #            'on_track_properties_activate':self.controller.on_track_properties,
            }
        )

        event.add_ui_callback(
            self.on_playback_end, 'playback_player_end', player.PLAYER
        )
        event.add_ui_callback(self.on_playback_end, 'playback_error', player.PLAYER)
        event.add_ui_callback(
            self.on_playback_start, 'playback_track_start', player.PLAYER
        )
        event.add_ui_callback(
            self.on_toggle_pause, 'playback_toggle_pause', player.PLAYER
        )
        event.add_ui_callback(self.on_track_tags_changed, 'track_tags_changed')
        event.add_ui_callback(self.on_buffering, 'playback_buffering', player.PLAYER)
        event.add_ui_callback(self.on_playback_error, 'playback_error', player.PLAYER)

        event.add_ui_callback(self.on_playlist_tracks_added, 'playlist_tracks_added')
        event.add_ui_callback(
            self.on_playlist_tracks_removed, 'playlist_tracks_removed'
        )

        # Settings
        self._on_option_set('gui_option_set', settings, 'gui/show_info_area')
        self._on_option_set('gui_option_set', settings, 'gui/show_info_area_covers')
        self._on_option_set('gui_option_set', settings, 'gui/show_status_bar')
        event.add_ui_callback(self._on_option_set, 'option_set')

    def _connect_panel_events(self):
        """
        Sets up panel events
        """

        # When there's nothing in the notebook, hide it
        self.controller.panel_notebook.connect(
            'page-added', self.on_panel_notebook_add_page
        )
        self.controller.panel_notebook.connect(
            'page-removed', self.on_panel_notebook_remove_page
        )

        # panels
        panels = self.controller.panel_notebook.panels

        for panel_name in ('playlists', 'radio', 'files', 'collection'):
            panel = panels[panel_name].panel
            do_sort = False

            if panel_name in ('files', 'collection'):
                do_sort = True

            panel.connect(
                'append-items',
                lambda panel, items, force_play: self.on_append_items(
                    items, force_play, sort=do_sort
                ),
            )
            panel.connect(
                'queue-items',
                lambda panel, items: self.on_append_items(
                    items, queue=True, sort=do_sort
                ),
            )
            panel.connect(
                'replace-items',
                lambda panel, items: self.on_append_items(
                    items, replace=True, sort=do_sort
                ),
            )

        ## Collection Panel
        panel = panels['collection'].panel
        panel.connect('collection-tree-loaded', self.on_collection_tree_loaded)

        ## Playlist Panel
        panel = panels['playlists'].panel
        panel.connect(
            'playlist-selected',
            lambda panel, playlist: self.playlist_container.create_tab_from_playlist(
                playlist
            ),
        )

        ## Radio Panel
        panel = panels['radio'].panel
        panel.connect(
            'playlist-selected',
            lambda panel, playlist: self.playlist_container.create_tab_from_playlist(
                playlist
            ),
        )

        ## Files Panel
        # panel = panels['files']

    def _update_alpha(self):
        if not settings.get_option('gui/use_alpha', False):
            return
        opac = 1.0 - float(settings.get_option('gui/transparency', 0.3))
        Gtk.Widget.set_opacity(self.window, opac)

    def _update_dark_hint(self):
        gs = Gtk.Settings.get_default()

        # We should use reset_property, but that's only available in > 3.20...
        if not hasattr(self, '_default_dark_hint'):
            self._default_dark_hint = gs.props.gtk_application_prefer_dark_theme

        if settings.get_option('gui/gtk_dark_hint', False):
            gs.props.gtk_application_prefer_dark_theme = True

        elif gs.props.gtk_application_prefer_dark_theme != self._default_dark_hint:
            # don't set it explicitly otherwise the app will revert to a light
            # theme -- what we actually want is to leave it up to the OS
            gs.props.gtk_application_prefer_dark_theme = self._default_dark_hint

    def do_get_property(self, prop):
        if prop.name == 'is-fullscreen':
            return self._fullscreen
        else:
            return GObject.Object.do_get_property(self, prop)

    def do_set_property(self, prop, value):
        if prop.name == 'is-fullscreen':
            if value:
                self.window.fullscreen()
            else:
                self.window.unfullscreen()
        else:
            GObject.Object.do_set_property(self, prop, value)

    def on_screen_changed(self, widget, event):
        """
        Updates the colormap on screen change
        """
        screen = widget.get_screen()
        visual = screen.get_rgba_visual() or screen.get_rgb_visual()
        self.window.set_visual(visual)

    def on_panel_notebook_add_page(self, notebook, page, page_num):
        if self.splitter.get_child1() is None:
            self.splitter.pack1(self.controller.panel_notebook)
            self.controller.panel_notebook.get_parent().child_set_property(
                self.controller.panel_notebook, 'shrink', False
            )

    def on_panel_notebook_remove_page(self, notebook, page, page_num):
        if notebook.get_n_pages() == 0:
            self.splitter.remove(self.controller.panel_notebook)

    def on_stop_button_motion_notify_event(self, widget, event):
        """
        Sets the hover state and shows SPAT icon
        """
        widget.__hovered = True
        if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
            widget.set_image(self.spat_image)
        else:
            widget.set_image(self.stop_image)

    def on_stop_button_leave_notify_event(self, widget, event):
        """
        Unsets the hover state and resets the button icon
        """
        widget.__hovered = False
        if not widget.is_focus() and ~(event.get_state() & Gdk.ModifierType.SHIFT_MASK):
            widget.set_image(self.stop_image)

    def on_stop_button_key_press_event(self, widget, event):
        """
        Shows SPAT icon on Shift key press
        """
        if event.keyval in (Gdk.KEY_Shift_L, Gdk.KEY_Shift_R):
            widget.set_image(
                Gtk.Image.new_from_icon_name('process-stop', Gtk.IconSize.BUTTON)
            )
            widget.toggle_spat = True

        if event.keyval in (Gdk.KEY_space, Gdk.KEY_Return):
            if widget.toggle_spat:
                self.on_spat_clicked()
            else:
                player.PLAYER.stop()

    def on_stop_button_key_release_event(self, widget, event):
        """
        Resets the button icon
        """
        if event.keyval in (Gdk.KEY_Shift_L, Gdk.KEY_Shift_R):
            widget.set_image(self.stop_image)
            widget.toggle_spat = False

    def on_stop_button_focus_out_event(self, widget, event):
        """
        Resets the button icon unless
        the button is still hovered
        """
        if not getattr(widget, '__hovered', False):
            widget.set_image(self.stop_image)

    def on_stop_button_press_event(self, widget, event):
        """
        Called when the user clicks on the stop button
        """
        if event.button == Gdk.BUTTON_PRIMARY:
            if event.get_state() & Gdk.ModifierType.SHIFT_MASK:
                self.on_spat_clicked()
        elif event.triggers_context_menu():
            m = menu.Menu(self)
            m.attach_to_widget(widget)
            m.add_simple(
                _("Toggle: Stop after Selected Track"),
                self.on_spat_clicked,
                'process-stop',
            )
            m.popup(event)

    def on_stop_button_release_event(self, widget, event):
        """
        Called when the user releases the mouse from the stop button
        """
        rect = widget.get_allocation()
        if 0 <= event.x < rect.width and 0 <= event.y < rect.height:
            player.PLAYER.stop()

    def on_stop_button_drag_motion(self, widget, context, x, y, time):
        """
        Indicates possible SPAT during drag motion of tracks
        """
        target = widget.drag_dest_find_target(context, None).name()
        if target == 'exaile-index-list':
            widget.set_image(self.spat_image)

    def on_stop_button_drag_leave(self, widget, context, time):
        """
        Resets the stop button
        """
        widget.set_image(self.stop_image)

    def on_stop_button_drag_data_received(
        self, widget, context, x, y, selection, info, time
    ):
        """
        Allows for triggering the SPAT feature
        by dropping tracks on the stop button
        """
        source_widget = Gtk.drag_get_source_widget(context)

        if selection.target.name() == 'exaile-index-list' and isinstance(
            source_widget, PlaylistView
        ):
            position = int(selection.data.split(',')[0])

            if position == source_widget.playlist.spat_position:
                position = -1

            source_widget.playlist.spat_position = position
            source_widget.queue_draw()

    def on_spat_clicked(self, *e):
        """
        Called when the user clicks on the SPAT item
        """
        trs = self.get_selected_page().view.get_selected_items()
        if not trs:
            return

        # TODO: this works, but implement this some other way in the future
        if player.QUEUE.current_playlist.spat_position == -1:
            player.QUEUE.current_playlist.spat_position = trs[0][0]
        else:
            player.QUEUE.current_playlist.spat_position = -1

        self.get_selected_page().view.queue_draw()

    def on_append_items(
        self, tracks, force_play=False, queue=False, sort=False, replace=False
    ):
        """
        Called when a panel (or other component)
        has tracks to append and possibly queue

        :param tracks: The tracks to append
        :param force_play: Force playing the first track if there
                            is no track currently playing. Otherwise
                            check a setting to determine whether the
                            track should be played
        :param queue: Additionally queue tracks
        :param sort: Sort before adding
        :param replace: Clear playlist before adding
        """
        if len(tracks) == 0:
            return

        page = self.get_selected_page()

        if sort:
            tracks = trax.sort_tracks(common.BASE_SORT_TAGS, tracks)

        if replace:
            page.playlist.clear()

        offset = len(page.playlist)
        page.playlist.extend(tracks)

        # extending the queue automatically starts playback
        if queue:
            if player.QUEUE is not page.playlist:
                player.QUEUE.extend(tracks)

        elif (
            force_play
            or settings.get_option('playlist/append_menu_starts_playback', False)
        ) and not player.PLAYER.current:
            page.view.play_track_at(offset, tracks[0])

    def on_playback_error(self, type, player, message):
        """
        Called when there has been a playback error
        """
        self.message.show_error(_('Playback error encountered!'), message)

    def on_buffering(self, type, player, percent):
        """
        Called when a stream is buffering
        """
        percent = min(percent, 100)
        self.statusbar.set_status(_("Buffering: %d%%...") % percent, 1)

    def on_track_tags_changed(self, type, track, tags):
        """
        Called when tags are changed
        """
        if track is player.PLAYER.current:
            self._update_track_information()

    def on_collection_tree_loaded(self, tree):
        """
        Updates information on collection tree load
        """
        self.statusbar.update_info()

    def on_exaile_loaded(self, event_type, exaile, nothing):
        """
        Updates information on exaile load
        """
        self.statusbar.update_info()
        event.remove_callback(self.on_exaile_loaded, 'exaile_loaded')

    def on_playlist_tracks_added(self, type, playlist, tracks):
        """
        Updates information on track add
        """
        self.statusbar.update_info()

    def on_playlist_tracks_removed(self, type, playlist, tracks):
        """
        Updates information on track removal
        """
        self.statusbar.update_info()

    def on_toggle_pause(self, type, player, object):
        """
        Called when the user clicks the play button after playback has
        already begun
        """
        if player.is_paused():
            image = self.play_image
            tooltip = _('Continue Playback')
        else:
            image = self.pause_image
            tooltip = _('Pause Playback')

        self.playpause_button.set_image(image)
        self.playpause_button.set_tooltip_text(tooltip)
        self._update_track_information()

    def on_playlist_container_switch_page(self, notebook, page, page_num):
        """
        Updates info after notebook page switch
        """
        page = notebook.get_nth_page(page_num)
        selection = page.view.get_selection()
        selection.connect('changed', self.on_playlist_view_selection_changed)
        self.statusbar.update_info()

    def on_playlist_view_selection_changed(self, selection):
        """
        Updates info after playlist page selection change
        """
        self.statusbar.update_info()

    def on_panel_filter_focus(self, *e):
        """
        Gives focus to the filter field of the current panel
        """
        try:
            self.controller.get_active_panel().filter.grab_focus()
        except (AttributeError, KeyError):
            pass

    def on_search_playlist_focus(self, *e):
        """
        Gives focus to the playlist search bar
        """
        plpage = get_selected_playlist()
        if plpage:
            plpage.get_search_entry().grab_focus()

    def on_save_playlist(self, *e):
        """
        Called when the user presses Ctrl+S
        """
        page = self.get_selected_playlist()
        if page:
            page.on_save()

    def on_save_playlist_as(self, *e):
        """
        Called when the user presses Ctrl+S
        Spawns the save as dialog of the current playlist tab
        """
        page = self.get_selected_playlist()
        if page:
            page.on_saveas()

    def on_clear_playlist(self, *e):
        """
        Clears the current playlist tab
        """
        page = self.get_selected_page()
        if page:
            page.playlist.clear()

    def on_open_item_activate(self, menuitem):
        """
        Shows a dialog to open media
        """

        def on_uris_selected(dialog, uris):
            uris.reverse()

            if len(uris) > 0:
                self.controller.open_uri(uris.pop(), play=True)

            for uri in uris:
                self.controller.open_uri(uri, play=False)

        dialog = dialogs.MediaOpenDialog(self.window)
        dialog.connect('uris-selected', on_uris_selected)
        dialog.show()

    def on_open_url_item_activate(self, menuitem):
        """
        Shows a dialog to open an URI
        """

        def on_uri_selected(dialog, uri):
            self.controller.open_uri(uri, play=False)

        dialog = dialogs.URIOpenDialog(self.window)
        dialog.connect('uri-selected', on_uri_selected)
        dialog.show()

    def on_open_directories_item_activate(self, menuitem):
        """
        Shows a dialog to open directories
        """

        def on_uris_selected(dialog, uris):
            uris.reverse()

            if len(uris) > 0:
                self.controller.open_uri(uris.pop(), play=True)

            for uri in uris:
                self.controller.open_uri(uri, play=False)

        dialog = dialogs.DirectoryOpenDialog(self.window)
        # Selecting empty folders is useless
        dialog.props.create_folders = False
        dialog.connect('uris-selected', on_uris_selected)
        dialog.show()

    def on_export_current_playlist_activate(self, menuitem):
        """
        Shows a dialog to export the current playlist
        """
        page = self.get_selected_page()

        if not page or not isinstance(page, PlaylistPage):
            return

        def on_message(dialog, message_type, message):
            """
            Show messages in the main window message area
            """
            if message_type == Gtk.MessageType.INFO:
                self.message.show_info(markup=message)
            elif message_type == Gtk.MessageType.ERROR:
                self.message.show_error(_('Playlist export failed!'), message)

            return True

        dialog = dialogs.PlaylistExportDialog(page.playlist, self.window)
        dialog.connect('message', on_message)
        dialog.show()

    def on_playlist_utilities_bar_visible_toggled(self, checkmenuitem):
        """
        Shows or hides the playlist utilities bar
        """
        settings.set_option(
            'gui/playlist_utilities_bar_visible', checkmenuitem.get_active()
        )

    def on_show_playing_track_item_activate(self, menuitem):
        """
        Tries to show the currently playing track
        """
        self.playlist_container.show_current_track()

    def on_about_item_activate(self, menuitem):
        """
        Shows the about dialog
        """
        dialog = dialogs.AboutDialog(self.window)
        dialog.show()

    def on_playback_start(self, type, player, object):
        """
        Called when playback starts
        Sets the currently playing track visible in the currently selected
        playlist if the user has chosen this setting
        """
        self._update_track_information()
        self.playpause_button.set_image(self.pause_image)
        self.playpause_button.set_tooltip_text(_('Pause Playback'))

    def on_playback_end(self, type, player, object):
        """
        Called when playback ends
        """
        self.window.set_title('Exaile')

        self.playpause_button.set_image(self.play_image)
        self.playpause_button.set_tooltip_text(_('Start Playback'))

    def _on_option_set(self, name, object, option):
        """
        Handles changes of settings
        """
        if option == 'gui/main_window_title_format':
            self.title_formatter.props.format = settings.get_option(
                option, self.title_formatter.props.format
            )

        elif option == 'gui/use_tray':
            usetray = settings.get_option(option, False)
            if self.controller.tray_icon and not usetray:
                self.controller.tray_icon.destroy()
                self.controller.tray_icon = None
            elif not self.controller.tray_icon and usetray:
                self.controller.tray_icon = tray.TrayIcon(self)

        elif option == 'gui/show_info_area':
            self.info_area.set_no_show_all(False)
            if settings.get_option(option, True):
                self.info_area.show_all()
            else:
                self.info_area.hide()
            self.info_area.set_no_show_all(True)

        elif option == 'gui/show_info_area_covers':
            cover = self.info_area.cover
            cover.set_no_show_all(False)
            if settings.get_option(option, True):
                cover.show_all()
            else:
                cover.hide()
            cover.set_no_show_all(True)

        elif option == 'gui/transparency':
            self._update_alpha()

        elif option == 'gui/gtk_dark_hint':
            self._update_dark_hint()

        elif option == 'gui/show_status_bar':
            if not settings.get_option('gui/show_status_bar', True):
                self.statusbar.hide()
            else:
                self.statusbar.show()

    def _on_volume_key(self, is_up):
        diff = int(
            100 * settings.get_option('gui/volue_key_step_size', VOLUME_STEP_DEFAULT)
        )
        if not is_up:
            diff = -diff

        player.PLAYER.modify_volume(diff)
        return True

    def _on_seek_key(self, is_forward):
        diff = settings.get_option('gui/seek_key_step_size', SEEK_STEP_DEFAULT)
        if not is_forward:
            diff = -diff

        if player.PLAYER.current:
            player.PLAYER.modify_time(diff)
            self.progress_bar.update_progress()

        return True

    def _on_prev_tab_key(self, *e):
        self.playlist_container.get_current_notebook().select_prev_tab()
        return True

    def _on_next_tab_key(self, *e):
        self.playlist_container.get_current_notebook().select_next_tab()
        return True

    def _on_playpause_button(self, *e):
        self.playpause()
        return True

    def _on_focus_playlist_tab(self, tab_nr):
        self.playlist_container.get_current_notebook().focus_tab(tab_nr)
        return True

    def _on_focus_playlist_container(self, *_e):
        self.playlist_container.focus()
        return True

    def _update_track_information(self):
        """
        Sets track information
        """
        track = player.PLAYER.current

        if not track:
            return

        self.window.set_title(self.title_formatter.format(track))

    def playpause(self):
        """
        Pauses the playlist if it is playing, starts playing if it is
        paused. If stopped, try to start playing the next suitable track.
        """
        if player.PLAYER.is_paused() or player.PLAYER.is_playing():
            player.PLAYER.toggle_pause()
        else:
            pl = self.get_selected_page()
            player.QUEUE.set_current_playlist(pl.playlist)
            try:
                trackpath = pl.view.get_selected_paths()[0]
                pl.playlist.current_position = trackpath[0]
            except IndexError:
                pass
            player.QUEUE.play(track=pl.playlist.current)

    def _setup_position(self):
        """
        Sets up the position and sized based on the size the window was
        when it was last moved or resized
        """
        if settings.get_option('gui/mainw_maximized', False):
            self.window.maximize()

        width = settings.get_option('gui/mainw_width', 500)
        height = settings.get_option('gui/mainw_height', 475)
        x = settings.get_option('gui/mainw_x', 10)
        y = settings.get_option('gui/mainw_y', 10)

        self.window.move(x, y)
        self.window.resize(width, height)

        pos = settings.get_option('gui/mainw_sash_pos', 200)
        self.splitter.set_position(pos)

    def on_delete_event(self, *e):
        """
        Called when the user attempts to close the window
        """
        sash_pos = self.splitter.get_position()
        if sash_pos > 10:
            settings.set_option('gui/mainw_sash_pos', sash_pos)

        if settings.get_option('gui/use_tray', False) and settings.get_option(
            'gui/close_to_tray', False
        ):
            self.window.hide()
        else:
            self.quit()
        return True

    def quit(self, *e):
        """
        Quits Exaile
        """
        self.window.hide()
        GLib.idle_add(self.controller.exaile.quit)
        return True

    def on_restart_item_activate(self, menuitem):
        """
        Restarts Exaile
        """
        self.window.hide()
        GLib.idle_add(self.controller.exaile.quit, True)

    def toggle_visible(self, bringtofront=False):
        """
        Toggles visibility of the main window
        """
        toggle_handled = self.emit('main-visible-toggle')

        if not toggle_handled:
            if (
                bringtofront
                and self.window.is_active()
                or not bringtofront
                and self.window.get_property('visible')
            ):
                self.window.hide()
            else:
                # the ordering for deiconify/show matters -- if this gets
                # switched, then the minimization detection breaks
                self.window.deiconify()
                self.window.show()

    def configure_event(self, *e):
        """
        Called when the window is resized or moved
        """
        # Don't save window size if it is maximized or fullscreen.
        if settings.get_option('gui/mainw_maximized', False) or self._fullscreen:
            return False

        width, height = self.window.get_size()
        if [width, height] != [
            settings.get_option("gui/mainw_" + key, -1) for key in ["width", "height"]
        ]:
            settings.set_option('gui/mainw_height', height, save=False)
            settings.set_option('gui/mainw_width', width, save=False)
        x, y = self.window.get_position()
        if [x, y] != [
            settings.get_option("gui/mainw_" + key, -1) for key in ["x", "y"]
        ]:
            settings.set_option('gui/mainw_x', x, save=False)
            settings.set_option('gui/mainw_y', y, save=False)

        return False

    def window_state_change_event(self, window, event):
        """
        Saves the current maximized and fullscreen
        states and minimizes to tray if requested
        """
        if event.changed_mask & Gdk.WindowState.MAXIMIZED:
            settings.set_option(
                'gui/mainw_maximized',
                bool(event.new_window_state & Gdk.WindowState.MAXIMIZED),
            )
        if event.changed_mask & Gdk.WindowState.FULLSCREEN:
            self._fullscreen = bool(event.new_window_state & Gdk.WindowState.FULLSCREEN)
            self.notify('is-fullscreen')

        # detect minimization state changes
        prev_minimized = self.minimized

        if not self.minimized:
            if (
                event.changed_mask & Gdk.WindowState.ICONIFIED
                and not event.changed_mask & Gdk.WindowState.WITHDRAWN
                and event.new_window_state & Gdk.WindowState.ICONIFIED
                and not event.new_window_state & Gdk.WindowState.WITHDRAWN
                and not self.window_state & Gdk.WindowState.ICONIFIED
            ):
                self.minimized = True
        else:
            if (
                event.changed_mask & Gdk.WindowState.WITHDRAWN
                and not event.new_window_state & (Gdk.WindowState.WITHDRAWN)
            ):  # and \
                self.minimized = False

        # track this
        self.window_state = event.new_window_state

        if settings.get_option('gui/minimize_to_tray', False):
            # old code to detect minimization
            # -> it must have worked at some point, perhaps this is a GTK version
            # specific set of behaviors? Current code works now on 2.24.17

            # if wm_state is not None:
            #    if '_NET_WM_STATE_HIDDEN' in wm_state[2]:
            #        show tray
            #        window.hide
            # else
            #    destroy tray

            if self.minimized != prev_minimized and self.minimized is True:
                if (
                    not settings.get_option('gui/use_tray', False)
                    and self.controller.tray_icon is None
                ):
                    self.controller.tray_icon = tray.TrayIcon(self)

                window.hide()
            elif (
                not settings.get_option('gui/use_tray', False)
                and self.controller.tray_icon is not None
            ):
                self.controller.tray_icon.destroy()
                self.controller.tray_icon = None

        return False

    def get_selected_page(self):
        """
        Returns the currently displayed playlist notebook page
        """
        return self.playlist_container.get_current_tab()

    def get_selected_playlist(self):
        try:
            page = self.get_selected_page()
        except AttributeError:
            return None
        if not isinstance(page, PlaylistPage):
            return None
        return page


class MainWindowTrackInfoPane(info.TrackInfoPane, providers.ProviderHandler):
    """
    Extends the regular track info pane by an area for custom widgets

    The mainwindow-info-area-widget provider is used to show widgets
    on the right of the info area. They should be small. The registered
    provider should provide a method 'create_widget' that takes the info
    area instance as a parameter, and that returns a Gtk.Widget to be
    inserted into the widget_area of the info area, and an attribute
    'name' that will be used when removing the provider.
    """

    def __init__(self, player):
        info.TrackInfoPane.__init__(self, player)

        self.__player = player
        self.widget_area = Gtk.Box()

        self.get_child().pack_start(self.widget_area, False, False, 0)

        self.__widget_area_widgets = {}

        # call this last if we're using simple_init=True
        providers.ProviderHandler.__init__(
            self, 'mainwindow-info-area-widget', target=player, simple_init=True
        )

    def get_player(self):
        """
        Retrieves the player object that this info area
        is associated with
        """
        return self._TrackInfoPane__player

    def on_provider_added(self, provider):
        name = provider.name
        widget = provider.create_widget(self)

        old_widget = self.__widget_area_widgets.get(name)
        if old_widget is not None:
            self.widget_area.remove(old_widget)
            old_widget.destroy()

        self.__widget_area_widgets[name] = widget
        self.widget_area.pack_start(widget, False, False, 0)
        widget.show_all()

    def on_provider_removed(self, provider):
        widget = self.__widget_area_widgets.pop(provider.name, None)
        if widget is not None:
            self.widget_area.remove(widget)
            widget.destroy()


class MainWindowStatusBarPane(Gtk.Statusbar, info.Statusbar, providers.ProviderHandler):
    """
    Extends the regular track info pane by an area for custom widgets

    The mainwindow-info-area-widget provider is used to show widgets
    on the right of the info area. They should be small. The registered
    provider should provide a method 'create_widget' that takes the info
    area instance as a parameter, and that returns a Gtk.Widget to be
    inserted into the widget_area of the info area, and an attribute
    'name' that will be used when removing the provider.
    """

    def __init__(self, player):
        Gtk.Statusbar.__init__(self)
        info.Statusbar.__init__(self, self)
        self.__player = player
        self.widget_area = Gtk.Box()

        self.set_halign(Gtk.Align.END)
        self.pack_start(self.widget_area, False, True, 0)
        self.reorder_child(self.widget_area, 0)

        self.__widget_area_widgets = {}

        # call this last if we're using simple_init=True
        providers.ProviderHandler.__init__(
            self, 'mainwindow-statusbar-widget', target=player, simple_init=True
        )

    def get_player(self):
        """
        Retrieves the player object that this info area
        is associated with
        """
        return self._TrackInfoPane__player

    def on_provider_added(self, provider):
        name = provider.name
        widget = provider.create_widget(self)

        old_widget = self.__widget_area_widgets.get(name)
        if old_widget is not None:
            self.widget_area.remove(old_widget)
            old_widget.destroy()

        self.__widget_area_widgets[name] = widget
        self.widget_area.pack_start(widget, False, True, 0)
        self.widget_area.show()

    def on_provider_removed(self, provider):
        widget = self.__widget_area_widgets.pop(provider.name, None)
        if widget is not None:
            self.widget_area.remove(widget)
            widget.destroy()


def get_playlist_container():
    return MainWindow._mainwindow.playlist_container


def get_playlist_notebook():
    '''Retrieves the primary playlist notebook'''
    return MainWindow._mainwindow.playlist_container.notebooks[0]


def get_selected_page():
    return MainWindow._mainwindow.get_selected_page()


def get_selected_playlist():
    return MainWindow._mainwindow.get_selected_playlist()


def mainwindow():
    return MainWindow._mainwindow


# vim: et sts=4 sw=4



# Copyright (C) 2012 Mathias Brodala
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from xl import settings


def migrate():
    """
    Enables the OSD plugin if the builtin OSD was originally enabled
    """
    plugins = settings.get_option('plugins/enabled', [])

    if settings.get_option('osd/enabled', False) and 'osd' not in plugins:
        settings.set_option('plugins/enabled', plugins + ['osd'])

    settings.set_option('osd/enabled', False)



# Copyright (C) 2008-2010 Adam Olsen
# Copyright (C) 2013-2015 Dustin Spicuzza
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

from gi.repository import GObject
from gi.repository import Gst

from xl import settings
from xl.nls import gettext as _

from . import gst_utils

import sys

import logging

logger = logging.getLogger(__name__)


__all__ = ['SINK_PRESETS', 'create_device', 'get_devices']


SINK_PRESETS = {
    # fmt: off
    "auto": {
        "name": _("Automatic")
    },
    "alsa": {
        "name": "ALSA",
        "pipe": "alsasink"
    },
    "oss": {
        "name": "OSS",
        "pipe": "osssink"
    },
    "pulse": {
        "name": "PulseAudio",
        "pipe": "pulsesink"
    },
    "jack": {
        "name": "JACK",
        "pipe": "jackaudiosink"
    },
    "directsoundsink": {
        'name': "DirectSound",
        'pipe': 'directsoundsink'
    },
    # TODO: uncomment when gstreamer wasapi plugin is stable
    #"wasapi": {
    #    'name': "WASAPI",
    #    'pipe': 'wasapisink'
    #},
    "custom": {
        "name": _("Custom")
    }
    # fmt: on
}


def __filter_presets():
    """Remove items in SINK_PRESETS that cannot be created."""
    for name, preset in list(SINK_PRESETS.items()):
        pipe = preset.get('pipe')
        if not pipe:
            continue  # No 'pipe' means this is auto or custom
        # In GStreamer < 1.28, Gst.ElementFactory.make always returns None when
        # the element can't be created.
        # In GStreamer >= 1.28, it can either return None or raise a custom
        # exception depending on whether the Gst Python override is installed.
        # See <https://gitlab.freedesktop.org/gstreamer/gstreamer/-/issues/4914>.
        try:
            if Gst.ElementFactory.make(pipe):
                continue
        except Exception:
            pass
        del SINK_PRESETS[name]


__filter_presets()


# Every audiosink is a bit different...
_autodetect_propnames = [
    'internal_name',  # pulseaudio
    'device_id',  # OSX
    'device_guid',  # directsound
    'device',  # wasapi
]


def _gst_device_autodetect():
    # Linux: support was added in gstreamer 1.4
    # OSX: support was added in gstreamer 1.10
    # Windows: support was added in gstreamer 1.14

    dm = Gst.DeviceMonitor.new()
    dm.add_filter('Audio/Sink', Gst.Caps.new_empty_simple('audio/x-raw'))

    show_all = settings.get_option("player/show_hidden_devices", False)
    dm.set_show_all_devices(show_all)

    dm.start()

    try:
        for device in dm.get_devices():
            for prop in _autodetect_propnames:
                if hasattr(device.props, prop):
                    device_id = getattr(device.props, prop)
                    break
            else:
                continue

            if not device_id:
                continue

            display_name = device.get_display_name()
            if hasattr(device.props, 'properties'):
                api = device.props.properties.get_string('device.api')
                if api:
                    display_name = '%s [%s]' % (display_name, api)

                    # TODO: remove this when wasapi is stable
                    if api == 'wasapi':
                        continue

            if show_all:
                device_cls = GObject.type_name_from_instance(device.g_type_instance)
                display_name = '%s (%s)' % (display_name, device_cls)

            yield (display_name, device_id, device.create_element)
    finally:
        dm.stop()


def get_devices():
    """
    Generator that yields (display_name, internal_name, fn(name)) where
    fn is a function that will create the audiosink when called
    """

    yield (
        _('Automatic'),
        'auto',
        lambda name: Gst.ElementFactory.make('autoaudiosink', name),
    )

    for info in _gst_device_autodetect():
        yield info


def create_device(player_name, return_errorsink=True):
    """
    Creates an audiosink based on the current settings. This will always
    return an audiosink, but sometimes it will return an audiosink that
    only sends error messages to the bus.

    ..note:: Only attempts to autoselect if the user has never specified a
             setting manually. Otherwise, they may be confused when it
             switches to a new output. For example, if they specified a USB
             device, and it is removed -- when restarting the program, they
             would not expect to automatically start playing on the builtin
             sound.
    """

    sink_type = settings.get_option('%s/audiosink' % player_name, 'auto')
    name = '%s-audiosink' % player_name
    sink = None
    errmsg = None

    if sink_type == 'auto':
        specified_device = settings.get_option(
            '%s/audiosink_device' % player_name, 'auto'
        )

        for _unused, device_id, create in get_devices():
            if specified_device == device_id:
                sink = create(name)
                break

        if sink is None:
            errmsg = _("Could not create audiosink (device: %s, type: %s)")
            errmsg = errmsg % (specified_device, sink_type)

    elif sink_type == 'custom':
        pipeline = settings.get_option("%s/custom_sink_pipe" % player_name, "")
        if not pipeline:
            errmsg = _("No custom pipeline specified!")
        else:
            try:
                sink = CustomAudioSink(pipeline, name)
            except Exception:
                errmsg = _("Error creating custom audiosink '%s'") % pipeline
                logger.exception(errmsg)

    else:
        preset = SINK_PRESETS.get(sink_type, None)
        if preset is None:
            errmsg = _("Invalid sink type '%s' specified") % sink_type
        else:
            sink = Gst.ElementFactory.make(preset['pipe'], name)
            if sink is None:
                errmsg = _("Could not create sink type '%s'") % preset['pipe']

    if errmsg is not None:
        logger.error(errmsg)
        if return_errorsink:
            sink = _get_error_audiosink(errmsg)

    return sink


def _get_error_audiosink(msg):
    sink = Gst.ElementFactory.make('fakesink', None)

    def handoff(s, b, p):
        s.message_full(
            Gst.MessageType.ERROR,
            Gst.stream_error_quark(),
            Gst.StreamError.FAILED,
            msg,
            msg,
            "",
            "",
            0,
        )

    sink.props.signal_handoffs = True
    sink.props.sync = True
    sink.connect('handoff', handoff)
    return sink


class CustomAudioSink(Gst.Bin):
    """
    A bin that holds the audio output sink element(s) for a custom
    defined sink
    """

    def __init__(self, pipeline, name):
        Gst.Bin.__init__(self, name='%s-audiosink' % name)

        elems = [Gst.parse_launch(elem.strip()) for elem in pipeline.split('!')]

        for e in elems:
            self.add(e)

        gst_utils.element_link_many(*elems)

        # GhostPad allows the bin to pretend to be an audio sink
        ghost = Gst.GhostPad.new("sink", elems[0].get_static_pad("sink"))
        self.add_pad(ghost)


if sys.platform == 'win32':
    from . import sink_windows

    priority_boost = sink_windows.get_priority_booster()

else:

    def priority_boost(player):
        # only needed on windows
        pass



# moodbar - Replace Exaile's seekbar with a moodbar
# Copyright (C) 2018  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os

from xl.nls import gettext as _
from xlgui.preferences import widgets

PLUGINDIR = os.path.dirname(os.path.realpath(__file__))

name = _('Moodbar')
ui = os.path.join(PLUGINDIR, "prefs.ui")


class UseWaveformPreference(widgets.CheckPreference):
    name = 'plugin/moodbar/use_waveform'
    default = False


class UseTintPreference(widgets.CheckPreference):
    name = 'plugin/moodbar/use_tint'
    default = False


class TintPreference(widgets.RGBAButtonPreference, widgets.CheckConditional):
    name = 'plugin/moodbar/tint'
    condition_preference_name = 'plugin/moodbar/use_tint'
    default = 'rgba(255, 255, 255, 0.2)'

    def __init__(self, preferences, widget):
        widgets.RGBAButtonPreference.__init__(self, preferences, widget)
        widgets.CheckConditional.__init__(self)


# vi: et sts=4 sw=4 tw=99



import json

from xl import event, settings
from xlgui.panel.collection import Order, DEFAULT_ORDERS

from . import cco_prefs


class CustomCollectionOrders:
    """
    Plugin adds user defined orders to the collection panel
    """

    collection_panel = None
    custom_orders = []
    default_orders_count = None

    def enable(self, exaile):
        """
        Called on startup of exaile
        """
        self.exaile = exaile
        event.add_callback(self._on_orders_update, 'cco_option_set')
        event.add_callback(self._on_active_order_update, 'gui_option_set')

    def _on_orders_update(self, event_name, event_source, option):
        if option != 'cco/orders':
            return
        self.populate_orders()

    def _on_active_order_update(self, event_name, event_source, option):
        """
        Called when 'gui/collection_active_view' is updated.
        """
        if option != 'gui/collection_active_view':
            return

        set = settings.get_option('gui/collection_active_view')
        if set > -1:
            # This is necessary because if on startup the last active view is not allready there
            # 'gui/collection_active_view' will be updated to -1
            settings.set_option('cco/collection_active_view', set)

    def disable(self, exaile):
        pass

    def on_exaile_loaded(self):
        """
        Called when the gui is loaded
        Before that there is no collection panel
        """
        self.collection_panel = self.exaile.gui.panel_notebook.panels[
            'collection'
        ].panel
        self.default_orders_count = len(DEFAULT_ORDERS)
        self.populate_orders()

    def populate_orders(self):
        setting = settings.get_option('cco/orders', None)
        if setting is None:
            return

        for order in self.custom_orders:
            self.collection_panel.orders.remove(order)
        self.custom_orders = []

        orders = json.loads(setting)

        for order in orders:
            levels = [s.strip() for s in order['levels'].split(',')]
            display = [s.strip() for s in order['display'].split(',')]

            final_sorting = display
            final_display = '$' + ' - $'.join(display)

            final = [final_sorting, final_display, final_sorting]
            levels.append(final)
            lvls = tuple(levels)

            new_order = Order(order['name'], lvls)
            self.collection_panel.orders.append(new_order)
            self.custom_orders.append(new_order)

        self.collection_panel.repopulate_choices()

    def get_preferences_pane(self):
        return cco_prefs


plugin_class = CustomCollectionOrders



# mpris2 - Support MPRIS 2 in Exaile
# Copyright (C) 2015-2016  Johannes Sasongko <sasongko@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.


import tempfile

from gi.repository import Gio, GLib

import xl.covers
import xl.event
import xl.player
import xl.settings

Variant = GLib.Variant


class MprisObject:
    def __init__(self, exaile, connection):
        self.exaile = exaile
        self.connection = connection
        self.cover_file = None
        self.signal_connections = []
        self.event_callbacks = callbacks = [
            (self._on_playback_track_start, 'playback_track_start', xl.player.PLAYER),
            (self._on_playback_track_end, 'playback_track_end', xl.player.PLAYER),
            (self._on_playback_toggle_pause, 'playback_toggle_pause', xl.player.PLAYER),
            (self._on_player_option_set, 'player_option_set'),
        ]
        for cb in callbacks:
            xl.event.add_ui_callback(*cb)
        if exaile.loading:
            xl.event.add_ui_callback(self._init_gui, 'exaile_loaded')
        else:
            self._init_gui()

    def __del__(self):
        self.destroy()

    def destroy(self):
        for obj, conn in self.signal_connections:
            obj.disconnect(conn)
        for cb in self.event_callbacks:
            xl.event.remove_callback(*cb)
        self.teardown()

    def _init_gui(self, *_args):
        if not hasattr(self.exaile.gui, 'main'):
            return  # No GUI
        conns = [
            (self.exaile.gui.main, 'notify::is-fullscreen', self._on_notify_fullscreen)
        ]
        self.signal_connections.extend(
            (obj, obj.connect(sig, handler)) for obj, sig, handler in conns
        )

    def teardown(self):
        """Quick destroy; just clean up our mess"""
        f = self.cover_file
        if f:
            f.close()

    def _new_cover_file(self):
        f = self.cover_file
        if f:
            f.close()
        self.cover_file = f = tempfile.NamedTemporaryFile(
            prefix='exaile.mpris2.cover.', suffix='.tmp'
        )
        uri = Gio.File.new_for_path(f.name).get_uri()
        return f, uri

    def _emit(self, interface, signame, *args):
        self.connection.emit_signal(
            None,
            '/org/mpris/MediaPlayer2',
            interface,
            signame,
            Variant.new_tuple(*args),
        )

    def _emit_propchange(self, interface, changed_props={}, invalidated_props=[]):
        self._emit(
            'org.freedesktop.DBus.Properties',
            'PropertiesChanged',
            Variant('s', interface),
            Variant('a{sv}', changed_props),
            Variant('as', invalidated_props),
        )

    def _get_metadata(self):
        track = xl.player.PLAYER.current
        if not track:
            return {}

        # TODO: Add more from https://www.freedesktop.org/wiki/Specifications/mpris-spec/metadata/

        # mpris

        meta = {'mpris:trackid': Variant('o', '/org/exaile/track/%d' % id(track))}
        v = track.get_tag_raw('__length')
        if v:
            meta['mpris:length'] = Variant('x', v * 1e6)
        v = xl.covers.MANAGER.get_cover(track, set_only=True, use_default=False)
        if v:
            f, uri = self._new_cover_file()
            f.write(v)
            f.flush()
            meta['mpris:artUrl'] = Variant('s', uri)

        # xesam

        v = track.get_tag_display('album')
        if v:
            meta['xesam:album'] = Variant('s', v)
        v = track.get_tag_display('artist', join=False)
        if v:
            meta['xesam:artist'] = Variant('as', v)
        v = track.get_tag_raw('genre')
        if v:
            # TODO: I've seen a client expect 'radio' on streams;
            # is that common usage?
            meta['xesam:genre'] = Variant('as', v)
        try:
            v = int(track.get_tag_display('tracknumber')[0])
        except (IndexError, TypeError, ValueError):
            pass
        else:
            meta['xesam:trackNumber'] = Variant('i', v)
        v = track.get_tag_display('title')
        if v:
            meta['xesam:title'] = Variant('s', v)
        v = track.get_loc_for_io()
        if v:
            meta['xesam:url'] = Variant('s', v)
        v = track.get_tag_raw('__rating')
        if v:
            meta['xesam:userRating'] = Variant('d', v / 100)

        return meta

    def _on_notify_fullscreen(self, obj, param):
        self._emit_propchange('org.mpris.MediaPlayer2', {'FullScreen': self.FullScreen})

    def _on_playback_track_start(self, event, player, track):
        self._emit_propchange(
            'org.mpris.MediaPlayer2.Player',
            {'Metadata': self.Metadata, 'PlaybackStatus': Variant('s', 'Playing')},
        )

    def _on_playback_track_end(self, event, player, track):
        self._emit_propchange(
            'org.mpris.MediaPlayer2.Player',
            {
                'Metadata': Variant('a{sv}', {}),
                'PlaybackStatus': Variant('s', 'Stopped'),
            },
        )

    def _on_playback_toggle_pause(self, event, player, track):
        self._emit_propchange(
            'org.mpris.MediaPlayer2.Player',
            {
                'PlaybackStatus': Variant(
                    's', 'Paused' if xl.player.PLAYER.is_paused() else 'Playing'
                )
            },
        )

    def _on_player_option_set(self, event, settings, option):
        if option == 'player/volume':
            self._emit_propchange(
                'org.mpris.MediaPlayer2.Player', {'Volume': self.Volume}
            )

    def _return_true(self):  # Positive reply for the "Can*" properties
        return Variant('b', True)

    # Root properties

    CanRaise = CanQuit = CanSetFullScreen = property(_return_true)

    @property
    def DesktopEntry(self):
        return Variant('s', 'exaile')

    @property
    def FullScreen(self):
        return Variant('b', self.exaile.gui.main.props.is_fullscreen)

    @FullScreen.setter
    def FullScreen(self, value):
        self.exaile.gui.main.props.is_fullscreen = value

    @property
    def HasTrackList(self):
        return Variant('b', False)

    @property
    def Identity(self):
        return Variant('s', 'Exaile')

    @property
    def SupportedMimeTypes(self):
        # Taken from exaile.desktop
        mimetypes = [
            'audio/musepack',
            'application/musepack',
            'application/x-ape',
            'audio/ape',
            'audio/x-ape',
            'audio/x-musepack',
            'application/x-musepack',
            'audio/x-mp3',
            'application/x-id3',
            'audio/mpeg',
            'audio/x-mpeg',
            'audio/x-mpeg-3',
            'audio/mpeg3',
            'audio/mp3',
            'audio/x-m4a',
            'audio/mpc',
            'audio/x-mpc',
            'audio/mp',
            'audio/x-mp',
            'application/ogg',
            'application/x-ogg',
            'audio/vorbis',
            'audio/x-vorbis',
            'audio/ogg',
            'audio/x-ogg',
            'audio/x-flac',
            'application/x-flac',
            'audio/flac',
        ]
        return Variant('as', mimetypes)

    @property
    def SupportedUriSchemes(self):
        # TODO: Call GstUriHandler.get_protocols on all GStreamer sources
        # and check if there are other useful protocols.
        return Variant('as', ['file', 'http', 'https', 'nfs', 'smb', 'sftp'])

    # Root methods

    def Raise(self):
        self.exaile.gui.main.window.present()

    def Quit(self):
        self.exaile.quit()

    # Player properties

    CanControl = CanGoNext = CanGoPrevious = CanPause = CanPlay = CanSeek = property(
        _return_true
    )

    @property
    def LoopStatus(self):
        playlist = xl.player.QUEUE.current_playlist
        state = playlist.get_repeat_mode()
        state_map = {'disabled': 'None', 'all': 'Playlist', 'track': 'Track'}
        assert state in state_map
        return Variant('s', state_map[state])

    @LoopStatus.setter
    def LoopStatus(self, value):
        value_s = value.get_string()
        state_map = {'None': 'disabled', 'Playlist': 'all', 'Track': 'track'}
        state = state_map.get(value_s)
        if not state:
            raise ValueError("invalid LoopStatus: %r" % value_s)
        playlist = xl.player.QUEUE.current_playlist
        playlist.set_repeat_mode(state)
        # TODO: Connect to actual event in playlist
        self._emit_propchange('org.mpris.MediaPlayer2.Player', {'LoopStatus': value})

    @property
    def MaximumRate(self):
        return Variant('d', 1)

    @property
    def Metadata(self):
        return Variant('a{sv}', self._get_metadata())

    @property
    def MinimumRate(self):
        return Variant('d', 1)

    @property
    def PlaybackStatus(self):
        state = xl.player.PLAYER.get_state()
        assert state in ('playing', 'paused', 'stopped')
        return Variant('s', state.capitalize())

    @property
    def Position(self):
        return Variant('x', xl.player.PLAYER.get_time() * 1e6)

    @property
    def Rate(self):
        return Variant('d', 1)

    @Rate.setter
    def Rate(self, value):
        pass

    @property
    def Shuffle(self):
        playlist = xl.player.QUEUE.current_playlist
        return Variant('b', playlist.get_shuffle_mode() != 'disabled')

    @Shuffle.setter
    def Shuffle(self, value):
        value_b = value.get_boolean()
        playlist = xl.player.QUEUE.current_playlist
        playlist.set_shuffle_mode('track' if value_b else 'disabled')
        # TODO: Connect to actual event in playlist
        self._emit_propchange(
            'org.mpris.MediaPlayer2.Player', {'Shuffle': Variant('b', value_b)}
        )

    @property
    def Volume(self):
        return Variant('d', xl.settings.get_option('player/volume', 1))

    @Volume.setter
    def Volume(self, value):
        xl.settings.set_option('player/volume', value.get_double())

    # Player methods

    def Next(self):
        if xl.player.PLAYER.is_playing():
            xl.player.QUEUE.next()

    def OpenUri(self, uri):
        pass  # TODO

    def Pause(self):
        xl.player.PLAYER.pause()

    def Play(self):
        player = xl.player.PLAYER
        state = player.get_state()
        if state == 'paused':
            player.unpause()
        elif state == 'stopped':
            queue = xl.player.QUEUE
            queue.play(queue.get_current())
        else:
            assert state == 'playing'  # Don't need to do anything

    def PlayPause(self):
        player = xl.player.PLAYER
        state = player.get_state()
        if state in ('playing', 'paused'):
            player.toggle_pause()
        else:
            assert state == 'stopped'
            queue = xl.player.QUEUE
            queue.play(queue.get_current())

    def Previous(self):
        if xl.player.PLAYER.is_playing():
            xl.player.QUEUE.prev()

    def Seek(self, offset):
        pass  # TODO

    def SetPosition(self, track_id, position):
        if track_id != '/org/exaile/track/%d' % id(xl.player.PLAYER.current):
            return
        xl.player.PLAYER.seek(position / 1e6)
        # TODO: Can we get this event from Exaile?
        self._emit(
            'org.mpris.MediaPlayer2.Player', 'Seeked', GLib.Variant('x', position)
        )

    def Stop(self):
        xl.player.PLAYER.stop()



"""
This plugin will make an IPC call to prevent the computer from suspending
during music playback

TODO: use Gtk.Application.inhibit() for less error prone inhibition.
"""

try:
    import _thread
except ImportError:
    import _dummy_thread as _thread
import logging
import os

import dbus
import subprocess

from xl.player import adapters, PLAYER

SUSPEND_PLUGIN = None
logger = logging.getLogger(__name__)


def enable(exaile):
    """
    Called when the plugin is enabled
    """
    global SUSPEND_PLUGIN

    if SUSPEND_PLUGIN is None:
        try:
            SUSPEND_PLUGIN = SuspendInhibit()
        except EnvironmentError:
            logger.error('Failed to Acquire Suspend Bus')
            raise
        except NotImplementedError:
            logger.error('Desktop Session not implemented')
            raise

        # allow plugin to finished enabling so that if user returns
        # to gnome plugin will not have to be re-enabled

    logger.info('Suspend Inhibitor Enabled')


def disable(exaile):
    """
    Called when the plugin is disabled
    """
    global SUSPEND_PLUGIN

    if SUSPEND_PLUGIN is not None:
        SUSPEND_PLUGIN.destroy()
        SUSPEND_PLUGIN = None

    logger.info('Suspend Inhibitor Disabled')


class SuspendInhibit:
    """
    Attempt to detect desktop session and initialize appropriate adapter
    """

    def __init__(self):
        # Attempt to detect Desktop Session Type
        session = os.getenv('DESKTOP_SESSION', '').lower()
        # see https://askubuntu.com/questions/72549/how-to-determine-which-window-manager-is-running
        xdg_session = os.getenv('XDG_CURRENT_DESKTOP', '').lower()

        # Attempt to find an adaptor that works
        if 'gnome' in session or 'gnome' in xdg_session:
            self.adapter = GnomeAdapter()
        elif 'kde' in session or 'kde' in xdg_session:
            try:
                self.adapter = PowerManagerAdapter()
            except EnvironmentError:
                # Fall back to powerdevil
                self.adapter = KdeAdapter()
        elif 'xfce' in session or 'xfce' in xdg_session:
            self.adapter = XfceAdapter()
        elif 'sway' in session or 'sway' in xdg_session:
            self.adapter = SwayAdapter()
        # TODO implement for LXDE, X-Cinnamon, Unity; systemd-inhibit
        elif session == '' and xdg_session == '':
            logger.warning('Could not detect Desktop Session, will try default \
                    Power Manager then Gnome')
            try:
                self.adapter = PowerManagerAdapter()
            except EnvironmentError:
                # Fall back to Gnome power manager
                self.adapter = GnomeAdapter()
        else:
            raise NotImplementedError(xdg_session)

    def destroy(self):
        self.adapter.destroy()


class SuspendAdapter(adapters.PlaybackAdapter):
    """
    Base class for Desktop Session suspend inhibitors

    Subclasses will have to override the DBus call methods

    Thread safe
    """

    PROGRAM = 'exaile'
    ACTIVITY = 'playing-music'

    def __init__(self):
        self.inhibited = False
        self.lock = _thread.allocate_lock()

        # Initialize parent object
        super().__init__(PLAYER)

        # Inhibit if player currently playing
        if PLAYER.is_playing():
            self.inhibit()

    def inhibit(self):
        """
        Inhibit user session suspension.

        Make DBus call to inhibit session suspension if
        session suspension not already inhibited.

        If suspending already inhibited call does nothing.
        """

        with self.lock:
            if not self.inhibited:
                self._inhibit_call()
                self.inhibited = True
                logger.info('Inhibited Suspend')

    def uninhibit(self):
        """
        Uninhibit user session suspension.

        If suspending already uninhibited call does nothing.
        """

        with self.lock:
            if self.inhibited:
                self._uninhibit_call()
                self.inhibited = False
                logger.info('Uninhibited Suspend')

    def is_inhibited(self):
        """Inhibit Status"""
        return self.inhibited

    def destroy(self):
        """Cleanup"""
        # Make sure to uninhibit when exiting
        self.uninhibit()
        adapters.PlaybackAdapter.destroy(self)
        logger.debug('Adapter Destroyed')

    """
        Playback Adapter Callbacks
    """

    def on_playback_track_start(self, event, player, track):
        self.inhibit()

    def on_playback_player_end(self, event, player, track):
        self.uninhibit()

    def on_playback_toggle_pause(self, event, player, track):
        if player.is_playing():
            self.inhibit()
        else:
            self.uninhibit()

    def _inhibit_call(self):
        """
        Override, overriding method must set self.inhibited value
        Must not block
        """
        raise NotImplementedError('Method not Overridden')

    def _uninhibit_call(self):
        """
        Override, overriding method must set self.inhibited value
        Must not block
        """
        raise NotImplementedError('Method not Overridden')


class DbusSuspendAdapter(SuspendAdapter):
    def __init__(self, bus_name, object_path, interface):
        try:
            bus = dbus.SessionBus()
            obj = bus.get_object(bus_name, object_path)
            self.iface = dbus.Interface(obj, interface)
            logger.info('Suspend Bus Acquired')
        except dbus.DBusException:
            raise EnvironmentError(bus_name + ' bus not available')

        super().__init__()

    def uninhibit(self):
        """
        Uninhibit user session suspension.

        Make DBus call to uninhibit session suspension if
        session suspension not already uninhibited.

        If suspending already uninhibited call does nothing.
        """

        with self.lock:
            if self.inhibited:
                if self.cookie is not None:
                    self._uninhibit_call()
                    self.cookie = None
                    self.inhibited = False
                    logger.info('Unihibited Suspend')
                else:
                    logger.error('Cannot Uninhibit Suspend without cookie')


class PowerManagerAdapter(DbusSuspendAdapter):
    """
    Default Adapter, implemented by most desktop sessions
    Adapter for org.freedesktop.PowerManagement.Inhibit Interface
    Some desktop sessions use different bus names for this interface
    and have other small variances
    """

    def __init__(
        self,
        bus_name='org.freedesktop.PowerManagement',
        object_name='/org/freedesktop/PowerManagement/Inhibit',
        interface_name='org.freedesktop.PowerManagement.Inhibit',
    ):
        super().__init__(bus_name, object_name, interface_name)

    def _dbus_inhibit_call(self):
        self.cookie = self.iface.Inhibit(self.PROGRAM, self.ACTIVITY)

    def _dbus_uninhibit_call(self):
        self.iface.UnInhibit(self.cookie)


class GnomeAdapter(DbusSuspendAdapter):
    """
    Gnome uses a similar interface to org.freedesktop.PowerManagement
    but is has different bus name, object path and interface name
    The inhibit and uninhibit method signatures are also different
    """

    SUSPEND_FLAG = 8

    def __init__(self):
        super().__init__(
            'org.gnome.SessionManager',
            '/org/gnome/SessionManager',
            'org.gnome.SessionManager',
        )

    def _dbus_inhibit_call(self):
        """
        Gnome Interface has more parameters
        """
        self.cookie = self.iface.Inhibit(
            self.PROGRAM, 1, self.ACTIVITY, self.SUSPEND_FLAG
        )

    def _dbus_uninhibit_call(self):
        """
        Gnome Interface has different case
        """
        self.iface.Uninhibit(self.cookie)


class KdeAdapter(PowerManagerAdapter):
    """
    Adapter for when org.freedesktop.PowerManager interface
    located at bus org.kde.powerdevil
    """

    def __init__(self):
        try:
            super().__init__()
        except EnvironmentError:
            # Fall back to other bus name
            super().__init__(bus_name='org.kde.powerdevil')


class XfceAdapter(PowerManagerAdapter):
    """
    Adapter for org.freedesktop.PowerManagement interface at bus name
    org.xfce.PowerManager
    """

    def __init__(self):
        try:
            super().__init__()
        except EnvironmentError:
            # Fall back to other bus name
            super().__init__(bus_name='org.xfce.PowerManager')


class SwayAdapter(SuspendAdapter):
    """
    Adapter for sway's IPC protocol/tool.
    """

    def _sway_inhibit_idle(self, status):
        subprocess.run(['swaymsg', '[app_id="exaile"]', 'inhibit_idle', status])

    def _inhibit_call(self):
        self._sway_inhibit_idle('open')

    def _uninhibit_call(self):
        self._sway_inhibit_idle('none')



#
# Copyright (C) 2015 Dustin Spicuzza <dustin@virtualroadside.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA
from os.path import abspath, join

import inspect
import warnings
import sys

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from xlgui import guiutil

__all__ = ['GtkTemplate']


class GtkTemplateWarning(UserWarning):
    pass


def _connect_func(builder, obj, signal_name, handler_name, connect_object, flags, cls):
    '''Handles GtkBuilder signal connect events'''

    if connect_object is None:
        extra = ()
    else:
        extra = (connect_object,)

    # The handler name refers to an attribute on the template instance,
    # so ask GtkBuilder for the template instance
    template_inst = builder.get_object(cls.__gtype_name__)

    if template_inst is None:  # This should never happen
        errmsg = (
            "Internal error: cannot find template instance! obj: %s; "
            "signal: %s; handler: %s; connect_obj: %s; class: %s"
            % (obj, signal_name, handler_name, connect_object, cls)
        )
        warnings.warn(errmsg, GtkTemplateWarning)
        return

    handler = getattr(template_inst, handler_name)

    if flags == GObject.ConnectFlags.AFTER:
        obj.connect_after(signal_name, handler, *extra)
    else:
        obj.connect(signal_name, handler, *extra)

    template_inst.__connected_template_signals__.add(handler_name)


def _register_template(cls, template_bytes):
    '''Registers the template for the widget and hooks init_template'''

    # This implementation won't work if there are nested templates, but
    # we can't do that anyways due to PyGObject limitations so it's ok

    cls.set_template(template_bytes)

    bound_methods = set()
    bound_widgets = set()

    # Walk the class, find marked callbacks and child attributes
    for name in dir(cls):
        o = getattr(cls, name, None)

        if inspect.ismethod(o):
            if hasattr(o, '_gtk_callback'):
                bound_methods.add(name)
                # Don't need to call this, as connect_func always gets called
                # cls.bind_template_callback_full(name, o)
        elif isinstance(o, _Child):
            cls.bind_template_child_full(name, True, 0)
            bound_widgets.add(name)

    # Have to setup a special connect function to connect at template init
    # because the methods are not bound yet
    cls.set_connect_func(_connect_func, cls)

    cls.__gtemplate_methods__ = bound_methods
    cls.__gtemplate_widgets__ = bound_widgets

    base_init_template = cls.init_template
    cls.init_template = lambda s: _init_template(s, cls, base_init_template)


def _init_template(self, cls, base_init_template):
    '''This would be better as an override for Gtk.Widget'''

    # TODO: could disallow using a metaclass.. but this is good enough
    # .. if you disagree, feel free to fix it and issue a PR :)
    if self.__class__ is not cls:
        raise TypeError(
            "Inheritance from classes with @GtkTemplate decorators "
            "is not allowed at this time"
        )

    connected_signals = set()
    self.__connected_template_signals__ = connected_signals

    base_init_template(self)

    for name in self.__gtemplate_widgets__:
        widget = self.get_template_child(cls, name)
        self.__dict__[name] = widget

        if widget is None:
            # Bug: if you bind a template child, and one of them was
            #      not present, then the whole template is broken (and
            #      it's not currently possible for us to know which
            #      one is broken either -- but the stderr should show
            #      something useful with a Gtk-CRITICAL message)
            raise AttributeError(
                "A missing child widget was set using "
                "GtkTemplate.Child and the entire "
                "template is now broken (widgets: %s)"
                % ', '.join(self.__gtemplate_widgets__)
            )

    for name in self.__gtemplate_methods__.difference(connected_signals):
        errmsg = (
            "Signal '%s' was declared with @GtkTemplate.Callback "
            + "but was not present in template"
        ) % name
        warnings.warn(errmsg, GtkTemplateWarning)


# TODO: Make it easier for IDE to introspect this
class _Child:
    """
    Assign this to an attribute in your class definition and it will
    be replaced with a widget defined in the UI file when init_template
    is called
    """

    __slots__ = []

    @staticmethod
    def widgets(count):
        '''
            Allows declaring multiple widgets with less typing::

                button    \
                label1    \
                label2    = GtkTemplate.Child.widgets(3)
        '''
        return [_Child() for _ in range(count)]


class _GtkTemplate:
    """
    Use this class decorator to signify that a class is a composite
    widget which will receive widgets and connect to signals as
    defined in a UI template. You must call init_template to
    cause the widgets/signals to be initialized from the template::

        @GtkTemplate(ui='foo.ui')
        class Foo(Gtk.Box):

            def __init__(self):
                super(Foo, self).__init__()
                self.init_template()

    The 'ui' parameter can either be a file path or a GResource resource
    path::

        @GtkTemplate(ui='/org/example/foo.ui')
        class Foo(Gtk.Box):
            pass

    To connect a signal to a method on your instance, do::

        @GtkTemplate.Callback
        def on_thing_happened(self, widget):
            pass

    To create a child attribute that is retrieved from your template,
    add this to your class definition::

        @GtkTemplate(ui='foo.ui')
        class Foo(Gtk.Box):

            widget = GtkTemplate.Child()


    Note: This is implemented as a class decorator, but if it were
    included with PyGI I suspect it might be better to do this
    in the GObject metaclass (or similar) so that init_template
    can be called automatically instead of forcing the user to do it.

    .. note:: Due to limitations in PyGObject, you may not inherit from
              python objects that use the GtkTemplate decorator.
    """

    __ui_path__ = None

    @staticmethod
    def Callback(f):
        """
        Decorator that designates a method to be attached to a signal from
        the template
        """
        f._gtk_callback = True
        return f

    Child = _Child

    @staticmethod
    def set_ui_path(*path):
        """
        If using file paths instead of resources, call this *before*
        loading anything that uses GtkTemplate, or it will fail to load
        your template file

        :param path: one or more path elements, will be joined together
                     to create the final path

        TODO: Alternatively, could wait until first class instantiation
              before registering templates? Would need a metaclass...
        """
        _GtkTemplate.__ui_path__ = abspath(join(*path))

    def __init__(self, ui):
        self.ui = ui

    def __call__(self, cls):
        if not issubclass(cls, Gtk.Widget):
            raise TypeError("Can only use @GtkTemplate on Widgets")

        # Nested templates don't work
        if hasattr(cls, '__gtemplate_methods__'):
            raise TypeError("Cannot nest template classes")

        # Load the template either from a resource path or a file
        # - Prefer the resource path first

        try:
            template_bytes = Gio.resources_lookup_data(
                self.ui, Gio.ResourceLookupFlags.NONE
            )
        except GLib.Error:
            ui = self.ui
            if isinstance(ui, (list, tuple)):
                ui = join(ui)

            if _GtkTemplate.__ui_path__ is not None:
                ui = join(_GtkTemplate.__ui_path__, ui)

            with open(ui, 'rb') as fp:
                template_bytes = GLib.Bytes.new(fp.read())

        if sys.platform == 'win32':
            string = template_bytes.get_data().decode("utf-8")
            template = guiutil.get_template_translated(string)
            template_bytes = template.encode('utf-8')
            template_bytes = GLib.Bytes.new(template_bytes)

        _register_template(cls, template_bytes)
        return cls


# Future shim support if this makes it into PyGI?
# if hasattr(Gtk, 'GtkTemplate'):
#    GtkTemplate = lambda c: c
# else:
GtkTemplate = _GtkTemplate



# Copyright (C) 2009-2010 Aren Olson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#
# The developers of the Exaile media player hereby grant permission
# for non-GPL compatible GStreamer and Exaile plugins to be used and
# distributed together with GStreamer and Exaile. This permission is
# above and beyond the permissions granted by the GPL license by which
# Exaile is covered. If you modify this code, you may extend this
# exception to your version of the code, but you are not obligated to
# do so. If you do not wish to do so, delete this exception statement
# from your version.

import dbus
import importlib
import logging
import os.path
import sys

from gi.repository import GLib

from xl.nls import gettext as _
from xl import providers, event, main
from xl.hal import Handler, UDisksProvider
from xl.devices import Device, KeyedDevice
from xl import playlist, trax, common, settings
from xl.trax import Track

from . import cdprefs, _cdguipanel
from sys import exc_info

logger = logging.getLogger(__name__)


if sys.platform.startswith('linux'):
    from . import linux_cd_parser


class CdPlugin:
    discid_parser = None
    musicbrainzngs_parser = None

    def __import_dependency(self, module_name):
        try:
            sys.path.append(os.path.dirname(__file__))
            full_name = module_name + '_parser'
            imported = importlib.import_module(full_name)
            sys.path.pop()
            return imported
        except ImportError:
            logger.warn(
                'Cannot import optional dependency "%s" for plugin cd.', module_name
            )
            logger.debug(
                'Traceback for importing dependency "%s" for plugin cd.',
                module_name,
                exc_info=True,
            )
            return None

    def enable(self, exaile):
        CdPlugin.discid_parser = self.__import_dependency('discid')
        CdPlugin.musicbrainzngs_parser = self.__import_dependency('musicbrainzngs')
        self.__exaile = exaile
        self.__udisks2 = None

    def on_exaile_loaded(self):
        # verify that hal/whatever is loaded, load correct provider
        if self.__exaile.udisks2 is not None:
            self.__udisks2 = UDisks2CdProvider()
            providers.register('udisks2', self.__udisks2)

    def teardown(self, exaile):
        if self.__udisks2 is not None:
            providers.unregister('udisks2', self.__udisks2)
        self.__udisks2 = None
        self.__exaile = None

    def disable(self, exaile):
        self.teardown(exaile)

    def get_preferences_pane(self):
        return cdprefs


plugin_class = CdPlugin


class _CDTrack:
    """
    @ivar track: Track number. Starts with 1, which is used for the TOC and contains data.
    @ivar data: `True` if this "track" contains data, `False` if it is audio
    @ivar minutes: Minutes from begin of CD
    @ivar seconds: Seconds after `minutes`, from begin of CD
    @ivar frames: Frames after `seconds`, from begin of CD
    """

    def __init__(self, entry):
        self.track, adrctrl, _format, addr = struct.unpack(TOC_ENTRY_FMT, entry)
        self.minutes, self.seconds, self.frames = struct.unpack(
            ADDR_FMT, struct.pack('i', addr)
        )

        # adr = adrctrl & 0xf
        ctrl = (adrctrl & 0xF0) >> 4

        self.data = False
        if ctrl & CDROM_DATA_TRACK:
            self.data = True

    def get_frame_count(self):
        return (self.minutes * 60 + self.seconds) * 75 + self.frames


class CDTocParser:
    # based on code from http://carey.geek.nz/code/python-cdrom/cdtoc.py

    def __init__(self, device):
        self.__raw_tracks = []
        self.__read_toc(device)

    def __read_toc(self, device):
        fd = os.open(device, os.O_RDONLY)
        try:
            toc_header = struct.pack(TOC_HEADER_FMT, 0, 0)
            toc_header = ioctl(fd, CDROMREADTOCHDR, toc_header)
            start, end = struct.unpack(TOC_HEADER_FMT, toc_header)

            # All tracks plus leadout
            for trnum in list(range(start, end + 1)) + [CDROM_LEADOUT]:
                entry = struct.pack(TOC_ENTRY_FMT, trnum, 0, CDROM_MSF, 0)
                entry = ioctl(fd, CDROMREADTOCENTRY, entry)
                self.__raw_tracks.append(_CDTrack(entry))
        finally:
            os.close(fd)

    def _get_track_lengths(self):
        """returns track length in seconds"""
        track = self.__raw_tracks[0]
        offset = track.get_frame_count()
        lengths = []
        for track in self.__raw_tracks[1:]:
            frame_end = track.get_frame_count()
            lengths.append((frame_end - offset) // 75)
            offset = frame_end
        return lengths


class CDPlaylist(playlist.Playlist):
    def __init__(self, name=_("Audio Disc"), device=None):
        playlist.Playlist.__init__(self, name=name)

        if not device:
            self.__device = "/dev/cdrom"
        else:
            self.__device = device
        self.__read_disc_index_async(device)

    @common.threaded
    def __read_disc_index_async(self, device):
        """This function must be run async because it does slow I/O"""
        logger.info('Starting to read disc index')

        if CdPlugin.discid_parser is not None:
            try:
                disc_id = CdPlugin.discid_parser.read_disc_id(device)
                logger.debug(
                    'Successfully read CD using discid with %i tracks. '
                    'Musicbrainz id: %s',
                    len(disc_id.tracks),
                    disc_id.id,
                )
                GLib.idle_add(self.__apply_disc_index, disc_id, None, None)
                return
            except Exception:
                logger.warn('Failed to read from cd using discid.', exc_info=True)

        if sys.platform.startswith('linux'):
            try:
                toc_entries, mcn = linux_cd_parser.read_cd_index(device)
                GLib.idle_add(self.__apply_disc_index, None, toc_entries, mcn)
                return
            except Exception:
                logger.warn('Failed to read metadata from CD.', exc_info=True)

        GLib.idle_add(self.__apply_disc_index, None, None, None)

    def __apply_disc_index(self, disc_id, toc_entries, mcn):
        """This function must be run sync because it accesses the track database"""
        logger.debug('Applying disc contents to playlist')
        if disc_id is not None:
            tracks = CdPlugin.discid_parser.parse_disc(disc_id, self.__device)
            if tracks is not None:
                allow_internet = settings.get_option(
                    'cd_import/fetch_metadata_from_internet', True
                )
                if allow_internet:
                    logger.info('Starting to get disc metadata')
                    self.__fetch_disc_metadata(disc_id, tracks)
        elif toc_entries is not None:
            tracks = linux_cd_parser.parse_tracks(toc_entries, mcn, self.__device)
        else:
            logger.error('Could not read disc index')
            tracks = None
        if tracks is not None:
            logger.debug('Read disc with tracks %s', tracks)
            self.extend(tracks)
        event.log_event('cd_info_retrieved', self, None)

    @common.threaded
    def __fetch_disc_metadata(self, disc_id, tracks):
        # TODO: show progress during work

        # TODO: Add more providers?
        # Discogs:
        #    Problem: Barely documented, no known support for disc_id
        # * https://github.com/discogs/discogs_client
        # * https://www.discogs.com/developers/
        # CDDB/freedb:
        # * old python code: http://pycddb.sourceforge.net/
        # * even older python code: http://cddb-py.sourceforge.net/
        # * http://ftp.freedb.org/pub/freedb/latest/DBFORMAT
        # * http://ftp.freedb.org/pub/freedb/latest/CDDBPROTO
        # * Servers: http://freedb.freedb.org/,
        if CdPlugin.musicbrainzngs_parser is not None:
            musicbrainz_data = CdPlugin.musicbrainzngs_parser.fetch_with_disc_id(
                disc_id
            )
            GLib.idle_add(
                self.__musicbrainz_metadata_fetched, musicbrainz_data, disc_id, tracks
            )

    def __musicbrainz_metadata_fetched(self, musicbrainz_data, disc_id, tracks):
        metadata = CdPlugin.musicbrainzngs_parser.parse(
            musicbrainz_data, disc_id, tracks
        )
        # TODO: progress: finished
        if metadata is not None:
            tracks, title = metadata
            logger.info('Finished getting disc metadata. Disc title: %s', title)
            event.log_event('cd_info_retrieved', self, title)
            self.name = title


class CDDevice(KeyedDevice):
    """
    represents a CD
    """

    class_autoconnect = True

    def __init__(self, dev):
        Device.__init__(self, dev)
        self.name = _("Audio Disc")
        self.dev = dev

    panel_type = _cdguipanel.CDPanel

    def __on_cd_info_retrieved(self, _event_type, cd_playlist, _disc_title):
        self.playlists.append(cd_playlist)
        self.connected = True

    def connect(self):
        if self.connected:
            return
        event.add_ui_callback(self.__on_cd_info_retrieved, 'cd_info_retrieved')
        CDPlaylist(device=self.dev)

    def disconnect(self):
        if not self.connected:
            return
        self.playlists = []
        self.connected = False
        CDDevice.destroy(self)


class UDisks2CdProvider(UDisksProvider):
    name = 'cd'
    PRIORITY = UDisksProvider.NORMAL

    def _get_num_tracks(self, obj, udisks):
        if obj.iface_type != 'org.freedesktop.UDisks2.Block':
            return

        try:
            drive = udisks.get_object_by_path(obj.props.Get('Drive'))
        except KeyError:
            return

        # Use number of audio tracks to identify supported media
        ntracks = drive.props.Get('OpticalNumAudioTracks')
        if ntracks > 0:
            return ntracks

    def get_priority(self, obj, udisks):
        ntracks = self._get_num_tracks(obj, udisks)
        if ntracks is not None:
            return self.PRIORITY

    def get_device(self, obj, udisks):
        device = obj.props.Get('Device', byte_arrays=True).decode('utf-8')
        device = device.strip('\0')
        return CDDevice(device)

    def on_device_changed(self, obj, udisks, device):
        if self._get_num_tracks(obj, udisks) is None:
            return 'remove'


# vim: et sts=4 sw=4
